# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import logging
import os
import sys
from logging import getLevelName

import numpy as np

################################################################################
# constants
################################################################################

#-----constants in SI units-----
GNewton    = 6.6743e-11
cLight     = 2.99792458e8
Msun       = 1.9884099021470415e+30
Mpc        = 3.085677581491367e+22
REarth     = 6.371e6
AU         = 1.4959787066e11
year       = 3.1536e7
hPlanck    = 6.62607015e-34
TEarthSol  = 86400   # solar day    (24h 60m 60s)
TEarthSid  = 86164.1 # sidereal day (23h 56m  4.0905s, found root 86164.09053826332)
TEarth     = TEarthSid
tc_offset  = 25134.55 # offset to tc: tc_gwbench = tc_bilby + tc_offset

#-----convert mass in solar masses to seconds-----
MTsun             = Msun * GNewton / cLight**3
#-----convert mass/distance in solar mass/Mpc to dimensionless-----
strain_fac        = GNewton / cLight**2 * Msun/Mpc
#-----convert seconds to radians with a periodicty of one day-----
time_to_rad_earth = 2 * np.pi / TEarth

################################################################################
# detector specifications --- location and geometry
################################################################################

# V-shaped interferometers have opening angles of pi/3, L-shaped interferometers have pi/2 (which is the default).
shape_opening_angles = {
    'L' : np.pi / 2.,
    'V1': np.pi / 3.,
    'V2': np.pi / 3.,
    'V3': np.pi / 3.,
    }

# V2 and V3 interferometers in a triangular configuration require a correction to the azimuth angle of the y-arm; default is 0.
# If the shape is not specified, the azimuth angle is not corrected and needs to be specified in the detector specifications correctly.
azimuth_shape_corr = {
    'L' : 0.,
    'V1': 0.,
    'V2': 2. * np.pi / 3.,
    'V3': 4. * np.pi / 3.,
    }

# the y-arm is the default arm for computation therefore a correction is needed, if either x-arm or bisector is specified; default is 0.
azimuth_arm_corr = {
    'x'       : np.pi / 2.,
    'bisector': np.pi / 4.,
    'y'       : 0.,
    }

# In-built detector specifications for the LIGO-Virgo-KAGRA detectors and various fiducial detector configurations.
# ----
# Switched to more precise angles for H, L, V, K, I, ET1, ET2, ET3 on 2022_03_29, found at: https://lscsoft.docs.ligo.org/lalsuite/lal/_l_a_l_detectors_8h_source.html
det_specs_built_in = {
    # 2G sites
    'H'     : {'longitude':-2.08405676917, 'latitude': 0.81079526383, 'arm_azimuth': np.pi-5.65487724844, 'which_arm':'y', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    'L'     : {'longitude':-1.58430937078, 'latitude': 0.53342313506, 'arm_azimuth': np.pi-4.40317772346, 'which_arm':'y', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    'V'     : {'longitude': 0.18333805213, 'latitude': 0.76151183984, 'arm_azimuth': np.pi-0.33916285222, 'which_arm':'y', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    'K'     : {'longitude': 2.396441015,   'latitude': 0.6355068497,  'arm_azimuth': np.pi-1.054113,      'which_arm':'y', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    'I'     : {'longitude': 1.33401332494, 'latitude': 0.24841853020, 'arm_azimuth': np.pi-1.57079637051, 'which_arm':'y', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    'LHO'   : {'longitude':-2.06982474503, 'latitude': 0.81079270401, 'arm_azimuth': 2.19910516123923,    'which_arm':'x', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    'LLO'   : {'longitude':-1.5572845695,  'latitude': 0.53342110078, 'arm_azimuth': 3.45080197126464,    'which_arm':'x', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    'LIO'   : {'longitude': 1.3444416672,  'latitude': 0.34231239582, 'arm_azimuth': 2.05248780779809,    'which_arm':'x', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    # fiducial ET sites
    'ET1'   : {'longitude': 0.18333805213, 'latitude': 0.76151183984, 'arm_azimuth': np.pi-0.33916285222, 'which_arm':'y', 'shape':'V1', 'radius':REarth, 'period':TEarthSid},
    'ET2'   : {'longitude': 0.18333805213, 'latitude': 0.76151183984, 'arm_azimuth': np.pi-0.33916285222, 'which_arm':'y', 'shape':'V2', 'radius':REarth, 'period':TEarthSid},
    'ET3'   : {'longitude': 0.18333805213, 'latitude': 0.76151183984, 'arm_azimuth': np.pi-0.33916285222, 'which_arm':'y', 'shape':'V3', 'radius':REarth, 'period':TEarthSid},
    'ETS1'  : {'longitude': 0.1643518379,  'latitude': 0.70714923527, 'arm_azimuth': 1.5707963267949,     'which_arm':'x', 'shape':'V1', 'radius':REarth, 'period':TEarthSid},
    'ETS2'  : {'longitude': 0.1643518379,  'latitude': 0.70714923527, 'arm_azimuth': 1.5707963267949,     'which_arm':'x', 'shape':'V2', 'radius':REarth, 'period':TEarthSid},
    'ETS3'  : {'longitude': 0.1643518379,  'latitude': 0.70714923527, 'arm_azimuth': 1.5707963267949,     'which_arm':'x', 'shape':'V3', 'radius':REarth, 'period':TEarthSid},
    # fiducial CE sites
    'C'     : {'longitude':-1.9691740,     'latitude': 0.764918,      'arm_azimuth': 0.,                  'which_arm':'y', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    'N'     : {'longitude':-1.8584265,     'latitude': 0.578751,      'arm_azimuth':-np.pi/3.,            'which_arm':'y', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    'S'     : {'longitude': 2.5307270,     'latitude':-0.593412,      'arm_azimuth': np.pi/4.,            'which_arm':'y', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    'CEA'   : {'longitude':-2.18166156499, 'latitude': 0.80285145592, 'arm_azimuth': 4.53785605518526,    'which_arm':'x', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    'CEB'   : {'longitude':-1.64060949687, 'latitude': 0.50614548308, 'arm_azimuth': 3.49065850398866,    'which_arm':'x', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    'CES'   : {'longitude': 2.5307270,     'latitude':-0.593412,      'arm_azimuth': np.pi/4.,            'which_arm':'y', 'shape':'L',  'radius':REarth, 'period':TEarthSid},
    }

def det_specs(loc, user_locs=None):
    '''
    Calculate the detector angles and shape for a given detector location.

    Parameters
    ----------
    loc : str
        Location (and implied orientation) of a detector.
    user_locs : dict
        User defined locations and orientations of detectors.

        This should specify
            'longitude':      longitude in radians,
            'latitude':       latitude in radians,
            'arm_azimuth':    azimuth of the specified arm in radians; the arm can be specified with 'which_arm' and the default 'y',
            'which_arm':      specify the arm with respect to which to take the azimuth: ['x', 'y', 'bisector'],
            'shape':          shape of the interferometer ['L', 'V1', 'V2', 'V3']; takes precededence over 'opening_angle',
            'opening_angle':  opening angle of the interferometer in radians; if 'shape' is specified, this is ignored,
            'radius':         Default: REarth;      radius from the geocenter of the detector's location in meters,
            'period':         Default: TEarthSid;   period of the detector around the geocenter in seconds,

    Returns
    -------
    alpha : float
        Longitude [rad]
    beta : float
        Latitude [rad]
    gamma : float
        Angle from 'Due East' to y-arm [rad]
    opening_angle or shape : float or str
        Opening angle of the detector or shape of interferometer (e.g. 'L', 'V1', ...), default is pi/2.
    radius : float
        Radius from the geocenter of the detector's location [m], default is REarth.
    period : float
        Period of the detector around the geocenter [s], default is TEarthSid.
    '''
    if user_locs is not None and loc in user_locs: _loc = user_locs[loc]
    elif loc in det_specs_built_in:                _loc = det_specs_built_in[loc]
    else: log_msg(f'det_specs: Provided location {loc} is neither among the implemented locations nor in provided user_locs.', level='ERROR')

    if 'shape' in _loc:
        if _loc['shape'] not in shape_opening_angles:
            log_msg(f'det_specs: Provided shape {_loc["shape"]} is not among the implemented shapes {list(shape_opening_angles.keys())}.', level='ERROR')
        opening_angle = shape_opening_angles[_loc['shape']]
        gamma_corr    = azimuth_shape_corr[_loc['shape']]
    else:
        opening_angle = _loc.get('opening_angle', np.pi / 2.)
        gamma_corr    = 0.

    return (_loc['longitude'],
            _loc['latitude'],
            _loc['arm_azimuth'] + azimuth_arm_corr[_loc['which_arm']] + gamma_corr, 
            opening_angle,
            _loc.get('radius', REarth),
            _loc.get('period', TEarthSid))

################################################################################
# basic functions
################################################################################

def reduce_symbols_strings(string1, string2):
    '''
    Combine two sympy symbols strings without duplicates.

    Parameters
    ----------
    string1 : str
        First sympy symbols string.
    string2 : str
        Second sympy symbols string.

    Returns
    -------
    str
        Combined sympy symbols string without duplicates.
    '''
    # combine the two lists
    symbols_list = string1.split(' ') + string2.split(' ')
    # remove duplicates
    symbols_list = list(dict.fromkeys(symbols_list))

    # recreate a string of symbols and return it
    return ' '.join(symbols_list)

def remove_symbols(string1, string2, keep_same=True):
    '''
    Remove symbols from one sympy symbols string that are not present in the other.

    Parameters
    ----------
    string1 : str
        First sympy symbols string.
    string2 : str
        Second sympy symbols string.
    keep_same : bool
        If True, keep the symbols that are present in both strings.

    Returns
    -------
    str
        Combined sympy symbols string without duplicates.
    '''
    symbs_list1 = string1.split(' ')
    symbs_list2 = string2.split(' ')
    # remove unwanted symbols from 1
    if keep_same:
        symbols_list = [x for x in symbs_list1 if x in symbs_list2]
    else:
        symbols_list = [x for x in symbs_list1 if x not in symbs_list2]

    # recreate a string of symbols and return it
    return ' '.join(symbols_list)

def get_sub_array_ids(arr, sub_arr):
    '''
    Get the indices of the elements of a subarray in a larger array.

    Parameters
    ----------
    arr : np.ndarray
        Larger array.
    sub_arr : np.ndarray
        Subarray.

    Returns
    -------
    np.ndarray
        Indices of the elements of the subarray in the larger array.
    '''
    return min_max_mask(arr, sub_arr[0], sub_arr[-1])

def get_sub_dict(dic, key_list, keep_in_dict=True):
    '''
    Get a subset of a dictionary based on a list of keys.

    Parameters
    ----------
    dic : dict
        Dictionary.
    key_list : list
        List of keys.
    keep_in_dict : bool
        If True, keep the keys in the dict. If False, keep the complientary keys.

    Returns
    -------
    dict
        Subset of the dictionary.
    '''
    if type(key_list) == str: key_list = key_list.split(' ')
    if keep_in_dict: return {k:v for k,v in dic.items() if k     in key_list}
    else:            return {k:v for k,v in dic.items() if k not in key_list}

def is_subset_lists(sub, sup):
    '''
    Check if a list is a subset of another list.

    Parameters
    ----------
    sub : list
        Sublist.
    sup : list
        Superlist.

    Returns
    -------
    bool
        True if the sublist is a subset of the superlist, False otherwise.
    '''
    return all([el in sup for el in sub])

def min_max_mask(arr, min_val=-np.inf, max_val=np.inf, strict_min=False, strict_max=False):
    '''
    Create a mask for an array based on minimum and maximum values.

    Parameters
    ----------
    arr : np.ndarray
        Array.
    min_val : float
        Minimum value. Default is -np.inf.
    max_val : float
        Maximum value. Default is np.inf.
    strict_min : bool
        If True, the minimum value is excluded. Default is False.
    strict_max : bool
        If True, the maximum value is excluded. Default is False.

    Returns
    -------
    mask : np.ndarray
        Mask for the array.
    '''
    if strict_min: min_mask = arr >  min_val
    else:          min_mask = arr >= min_val

    if strict_max: max_mask = arr <  max_val
    else:          max_mask = arr <= max_val

    return np.logical_and(min_mask, max_mask)

################################################################################
# waveform manipluations
################################################################################

#-----stable evaluation of exponential phase factors-----
def mod_1(val, np=np):
    '''
    Return stable evaluation of mod(val, 1) for a value val to be inserted in exp(1j * 2pi * val).
    The returned value is always in the range [0, 1).

    Parameters
    ----------
    val : np.ndarray or jnp.ndarray
        Values

    Returns
    -------
    mod : np.ndarray or jnp.ndarray
        mod(val, 1)
    '''
    return val - np.floor(val)

#-----convert waveform polarizations to amplitude and phase-----
def transform_hfpc_to_amp_pha(hfpc, f, params_list, np=np):
    '''
    Calculate and transform the complex frequency domain waveform for the plus and cross polarizations to amplitude and phase.

    Parameters
    ----------
    hfpc : np.ndarray
        Complex frequency domain waveform for the plus and cross polarizations.
    f : np.ndarray
        Frequency array.
    params_list : list
        List of parameters.

    Returns
    -------
    np.ndarray
        Amplitude of the plus polarization.
    np.ndarray
        Phase of the plus polarization.
    np.ndarray
        Amplitude of the cross polarization.
    np.ndarray
        Phase of the cross polarization.
    '''
    hfp, hfc = hfpc(f, *params_list)
    return pl_cr_to_amp_pha(hfp, hfc, np=np)

def pl_cr_to_amp_pha(hfp, hfc, np=np):
    '''
    Transform the plus and cross polarizations to amplitude and phase.

    Parameters
    ----------
    hfp : np.ndarray
        Plus polarization.
    hfc : np.ndarray
        Cross polarization.

    Returns
    -------
    np.ndarray
        Amplitude of the plus polarization.
    np.ndarray
        Phase of the plus polarization.
    np.ndarray
        Amplitude of the cross polarization.
    np.ndarray
        Phase of the cross polarization.
    '''
    hfp_amp, hfp_pha = amp_pha_from_z(hfp, np=np)
    hfc_amp, hfc_pha = amp_pha_from_z(hfc, np=np)
    return hfp_amp, hfp_pha, hfc_amp, hfc_pha

#-----convert amp/phase derivatives to re/im ones-----
def z_deriv_from_amp_pha(amp, pha, del_amp, del_pha, np=np):
    '''
    Calculate the real and imaginary part of waveform derivatives from the amplitude and phase and their derivatives.

    Parameters
    ----------
    amp : np.ndarray
        Amplitude of the waveform.
    pha : np.ndarray
        Phase of the waveform.
    del_amp : np.ndarray
        Derivative of the amplitude.
    del_pha : np.ndarray
        Derivative of the phase.

    Returns
    -------
    np.ndarray
        Real part of the waveform derivative.
    np.ndarray
        Imaginary part of the waveform derivative.
    '''
    if len(del_amp.shape) == 2: return np.array([ del_amp[:,i] * np.exp(1j*pha) + amp * np.exp(1j*pha) * 1j * del_pha[:,i]
                                                  for i in range(del_amp.shape[1]) ], dtype=complex).T
    else:                       return del_amp * np.exp(1j*pha) + amp * np.exp(1j*pha) * 1j * del_pha

#-----re/im vs. amp/phase transformations-----
def re_im_from_amp_pha(amp, pha, np=np):
    '''
    Calculate the real and imaginary part of a complex number from its amplitude and phase.

    Parameters
    ----------
    amp : np.ndarray
        Amplitude.
    pha : np.ndarray
        Phase.

    Returns
    -------
    np.ndarray
        Real part of the complex number.
    np.ndarray
        Imaginary part of the complex number.
    '''
    return re_im_from_z(z_from_amp_pha(amp, pha), np=np)

def amp_pha_from_re_im(re, im, np=np):
    '''
    Calculate the amplitude and phase of a complex number from its real and imaginary part.

    Parameters
    ----------
    re : np.ndarray
        Real part.
    im : np.ndarray
        Imaginary part.

    Returns
    -------
    np.ndarray
        Amplitude.
    np.ndarray
        Phase.
    '''
    return amp_pha_from_z(z_from_re_im(re, im), np=np)

#-----re/im or amp/phase vs. complex number transformations-----
def re_im_from_z(z, np=np):
    '''
    Calculate the real and imaginary part of a complex number.

    Parameters
    ----------
    z : np.ndarray
        Complex number.

    Returns
    -------
    np.ndarray
        Real part of the complex number.
    np.ndarray
        Imaginary part of the complex number.
    '''
    return np.real(z), np.imag(z)

def z_from_re_im(re, im):
    '''
    Calculate the complex number from its real and imaginary part.

    Parameters:
        re (array): Real part.
        im (array): Imaginary part.

    Returns:
        array: Complex number.
    '''
    return re + 1j * im

def amp_pha_from_z(z, np=np):
    '''
    Calculate the amplitude and phase of a complex number.

    Parameters
    ----------
    z : np.ndarray
        Complex number.

    Returns
    -------
    np.ndarray
        Amplitude.
    np.ndarray
        Phase.
    '''
    return np.abs(z), np.unwrap(np.angle(z))

def z_from_amp_pha(amp, pha, np=np):
    '''
    Calculate the complex number from its amplitude and phase.

    Parameters
    ----------
    amp : np.ndarray
        Amplitude.
    pha : np.ndarray
        Phase.

    Returns
    -------
    np.ndarray
        Complex number.
    '''
    return amp * np.exp(1j*pha)


################################################################################
# IO functions
################################################################################

#-----Block and unblock printing-----
def block_print(active=True):
    '''
    Block printing to the standard output.

    Parameters
    ----------
    active : bool
        If True, block the printing. If False, do nothing.
    '''
    if active: sys.stdout = open(os.devnull, 'w')
    return

def unblock_print(active=True):
    '''
    Unblock printing to the standard output.

    Parameters
    ----------
    active : bool
        If True, unblock the printing. If False, do nothing.
    '''
    if active: sys.stdout = sys.__stdout__
    return

#-----sending warning or error message-----
def log_msg(message, logger=None, level='INFO'):
    '''
    Log a message to the standard output or a logger.

    Parameters
    ----------
    message : str
        Message to be logged.
    logger : logging.Logger
        Logger to log the message. If None, the message is printed to the standard output.
    level : str
        Level of the message, if a logger is used. Can be 'DEBUG', 'INFO', 'WARNING', 'ERROR', or 'CRITICAL'.
    '''
    if logger is None: print(level + ': ' + message)
    else:              logger.log(getLevelName(level), message)
    if level in ['ERROR', 'CRITICAL']: sys.exit()

def get_logger(name, level='INFO', stdout=True, logfile=None):
    '''
    Get a logger.

    Parameters
    ----------
    name : str
        Name of the logger.
    level : str
        Level of the logger. Can be 'DEBUG', 'INFO', 'WARNING', 'ERROR', or 'CRITICAL'.
    stdout : bool
        If True, log to the standard output.
    logfile : str
        If not None, log to a file with the given name.

    Returns
    -------
    logging.Logger
        Logger.
    '''
    logging.basicConfig(format = '%(asctime)s - %(name)s - %(levelname)s : %(message)s')
    if stdout: logging.basicConfig(stream = sys.stdout)
    if logfile is not None: logging.basicConfig(filename = logfile, filemode = 'w')
    logger = logging.getLogger(name)
    set_logger_level(logger, level)
    return logger

def set_logger_level(logger, level):
    '''
    Set the level of a logger.

    Parameters
    ----------
    logger : logging.Logger
        Logger.
    level : str
        Level of the logger. Can be 'DEBUG', 'INFO', 'WARNING', 'ERROR', or 'CRITICAL'.
    '''
    logger.setLevel(level)


################################################################################
# network_spec handlers
################################################################################

def read_det_keys_from_label(network_label):
    '''
    Read the network label and find all detectors.

    Parameters
    ----------
    network_label : str
        Network label.

    Returns
    -------
    list
        list: List of detector keys.
    '''
    det_keys = []

    ##-----read the network label and find all detectors-----
    keys = list(network_label)

    # - in network list means that all 2G detectors up to that index are to be
    # taken at aLIGO sensitivity
    aLIGO = int('-' in keys)
    if aLIGO: aLIGO_id = keys.index('-')

    # + in network list means that all 2G detectors up to that index are to be
    # taken at A+ sensitivity
    a_pl = int('+' in keys)
    if a_pl: a_pl_id = keys.index('+')

    # v in network list means that all 2G detectors up to that index are to be
    # taken at Voyager sensitivity
    voy = int('v' in keys)
    if voy:
        voy_id = keys.index('v')
        tmp = int(keys[voy_id+1] == 'p')
        voy_pmo = tmp * 'PMO' + (1-tmp) * 'CBO'

    # find out which locations with which PSDs are in the network
    for loc in det_specs_built_in:
        if loc in keys:
            loc_id = keys.index(loc)

            if loc in ('H','L','V','K','I'):
                if aLIGO and loc_id < aLIGO_id:
                    name = 'aLIGO_'+loc
                elif a_pl and loc_id < a_pl_id:
                    if loc == 'V':
                        name = 'V+_'+loc
                    elif loc == 'K':
                        name = 'K+_'+loc
                    else:
                        name = 'A+_'+loc
                elif voy and loc_id < voy_id:
                    name = 'Voyager-{}_{}'.format(voy_pmo,loc)

            elif loc in ('C','N','S'):
                if keys[loc_id+1] == 'c':
                    name = f'CE-{keys[loc_id+2]}0'
                    if   keys[loc_id+2] == 'l': name += '-LF'
                    elif keys[loc_id+2] == 'p': name += '-PM'
                    name += f'_{loc}'
                else:
                    ce_a = int(keys[loc_id+1] == 'a') # 0 for i, 1 for a - CE1 as i, CE2 as a
                    ce_arm = int(keys[loc_id+2])*10  # arm length (n for n*10km)
                    tmp = int(keys[loc_id+3] == 'p')
                    ce_pmo = tmp * 'PMO' + (1-tmp) * 'CBO'
                    name = f'CE{ce_a+1}-{ce_arm}-{ce_pmo}_{loc}'

            det_keys.append(name)

    # add 3 ET detectors
    if 'E' in keys:
        for name in ['ET_ET1','ET_ET2','ET_ET3']:
            det_keys.append(name)

    return det_keys



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import numpy as np

from gwbench.antenna_pattern_np import ant_pat_vectors, calc_gra, calc_rotating_time, calc_time_delay, det_quants, loc_phase_func

cos = np.cos
sin = np.sin
pi  = np.pi

# detector response [DL, tc, phic, ra, dec, psi]
def detector_response_ana_derivs(f, params_dic, hf, hfp, hfc, Flp, loc, use_rot, user_locs=None):
    '''
    Return the paratial derivatives of the detector response with respect to the following parameters:
    - DL: the luminosity distance
    - tc: the time at coalescence
    - phic: the phase at coalescence
    - ra: the right ascension
    - dec: the declination
    - psi: the polarization angle

    Parameters
    ----------
    f : np.ndarray
        The frequency array.
    params_dic : dict
        The dictionary of the waveform parameters.
    hf : np.ndarray
        The detector response.
    hfp : np.ndarray
        The plus polarization of the waveform.
    hfc : np.ndarray
        The cross polarization of the waveform.
    Flp : np.ndarray
        The location phase factor.
    loc : np.ndarray
        The location of the detector.
    use_rot : bool
        Whether to use the rotation of the Earth or not.
    user_locs : dict, optional
        User-specified locations of the detectors.

    Returns
    -------
    dict
        The partial derivatives of the detector response.
    '''
    d_hf              = waveform_ana_derivs(f, params_dic, hf)
    d_Fp, d_Fc, d_Flp = antenna_pattern_and_loc_phase_fac_ana_derivs(f, params_dic, loc, use_rot, Flp=1, user_locs=user_locs)
    # d_Flp ~ Flp, hence in order to avoid unnecessary computations on arrays:
    # we pass Flp=1 and simply multiply with hf instead of hf/Flp

    return {
        'DL'    : d_hf['DL'],
        'phic'  : d_hf['phic'],
        'tc'    : d_hf['tc']  + Flp * (hfp * d_Fp['tc']  + hfc * d_Fc['tc'] ) + d_Flp['tc']  * hf,
        'ra'    :               Flp * (hfp * d_Fp['ra']  + hfc * d_Fc['ra'] ) + d_Flp['ra']  * hf,
        'dec'   :               Flp * (hfp * d_Fp['dec'] + hfc * d_Fc['dec']) + d_Flp['dec'] * hf,
        'psi'   :               Flp * (hfp * d_Fp['psi'] + hfc * d_Fc['psi']),
        }


# waveform (polarization or detector response) [DL, tc, phic]
def waveform_ana_derivs(f, params_dic, hf):
    '''
    Return the partial derivatives of the waveform (either a waveform polarization or the detector response)
    with respect to the following parameters:
    - DL: the luminosity distance
    - tc: the time at coalescence
    - phic: the phase at coalescence

    Parameters
    ----------
    f : np.ndarray
        The frequency array.
    params_dic : dict
        The dictionary of the waveform parameters.
    hf : np.ndarray
        Waveform (either a waveform polarization or the detector response).

    Returns
    -------
    dict
        The partial derivatives of the waveform.
    '''
    return {
        'DL'    : -hf / params_dic['DL'],
        'tc'    : (1j*2*pi) * f * hf,
        'phic'  : -1j * hf,
        }


# antenna pattern and location phase factor [ra, dec, psi, tc]
def antenna_pattern_and_loc_phase_fac_ana_derivs(f, params_dic, loc, use_rot, Flp=None, user_locs=None):
    '''
    Calculate the partial derivatives of the antenna pattern functions Fp, Fc and the location phase factor Flp
    with respect to the following parameters:
    - ra: the Greenwich right ascension
    - dec: the declination
    - psi: the polarization angle
    - tc: the time of coalescence

    Parameters
    ----------
    f : np.ndarray
        The frequency array [Hz].
    params_dic : dict
        The dictionary of the waveform parameters.
    loc : np.ndarray
        The location of the detector.
    use_rot : bool
        Whether to use the rotation of the Earth or not.
    Flp : np.ndarray, optional
        The location phase factor. Can be passed if available to avoid unnecessary computations.
    user_locs : dict
        User-specified locations of the detectors.

    Returns
    -------
    del_ra_dec_psi_tc_Fp : dict
        The partial derivatives of the plus polarization of the antenna pattern.
    del_ra_dec_psi_tc_Fc : dict
        The partial derivatives of the cross polarization of the antenna pattern.
    del_ra_dec_tc_Flp : dict
        The partial derivatives of the location phase factor.
    '''
    det_ten, det_vec, period_to_rad = det_quants(loc, user_locs=user_locs)
    time                            = calc_rotating_time(params_dic['tc'], f, params_dic['Mc'], params_dic['eta'], use_rot)
    gra_Flp                         = calc_gra(params_dic['ra'], time, period_to_rad)
    time_delay                      = calc_time_delay(gra_Flp, params_dic['dec'], det_vec)
    gra_Fpc                         = calc_gra(params_dic['ra'], time + time_delay, period_to_rad)

    del_gra_td, del_dec_td          = calc_d_time_delay(gra_Flp, params_dic['dec'], f, det_vec)
    del_ra_gra_Fpc                  =  (1 - period_to_rad * del_gra_td)
    del_dec_gra_Fpc                 = -period_to_rad *      del_dec_td
    del_tc_gra_Fpc                  = -period_to_rad * (1 + del_gra_td)

    del_gra_dec_psi_Fpc             = np.sum(np.matmul(
        np.einsum('ijkl->jlik', calc_d_ant_pat_funcs(det_ten, *ant_pat_vectors(gra_Fpc, params_dic['dec'], params_dic['psi']))),
        np.einsum('ijkl->ilkj', calc_d_ant_pat_vectors(gra_Fpc, params_dic['dec'], params_dic['psi'])) ), axis=0)

    if Flp is None: Flp             = loc_phase_func(f, time_delay)
    d_Flp_fac                       = (1j*2*pi) * f * Flp
    del_ra_Flp                      = d_Flp_fac * del_gra_td

    # return dictionaries of the partial derivatives of Fp, Fc, Flp
    return  {   'ra'  : del_gra_dec_psi_Fpc[:,0,0] * del_ra_gra_Fpc,
                'dec' : del_gra_dec_psi_Fpc[:,0,1] + del_gra_dec_psi_Fpc[:,0,0] * del_dec_gra_Fpc,
                'psi' : del_gra_dec_psi_Fpc[:,0,2],
                'tc'  : del_gra_dec_psi_Fpc[:,0,0] * del_tc_gra_Fpc,
                }, \
            {   'ra'  : del_gra_dec_psi_Fpc[:,1,0] * del_ra_gra_Fpc,
                'dec' : del_gra_dec_psi_Fpc[:,1,1] + del_gra_dec_psi_Fpc[:,1,0] * del_dec_gra_Fpc,
                'psi' : del_gra_dec_psi_Fpc[:,1,2],
                'tc'  : del_gra_dec_psi_Fpc[:,1,0] * del_tc_gra_Fpc,
                }, \
            {   'ra' : del_ra_Flp,
                'dec': d_Flp_fac * del_dec_td,
                'tc' : -period_to_rad * del_ra_Flp,
                }


# time delay [gra, dec]
def calc_d_time_delay(gra, dec, f, det_vec):
    '''
    Calculate the partial derivatives of the time delay with respect to the following parameters:
    - gra: the Greenwich right ascension
    - dec: the declination

    Parameters
    ----------
    gra : np.ndarray
        The Greenwhich right ascension.
    dec : np.ndarray
        The declination.
    f : np.ndarray
        The frequency array.
    det_vec : np.ndarray
        The location vector.

    Returns
    -------
    np.ndarray
        The partial derivatives of the time delay.
    '''
    # del_gra_time_delay, del_dec_time_delay
    return np.matmul(det_vec, np.array([ -sin(gra)*cos(dec),  cos(gra)*cos(dec),          np.zeros_like(gra) ])), \
           np.matmul(det_vec, np.array([ -cos(gra)*sin(dec), -sin(gra)*sin(dec), cos(dec)* np.ones_like(gra) ]))


# antenna patterns [XX, YY]
def calc_d_ant_pat_funcs(det_ten, XX, YY):
    '''
    Calculate the partial derivatives of the plus and cross polarizations of the antenna pattern with respect to the following parameters:
    - XX [array_like]: x-arm antenna pattern vector
    - YY [array_like]: y-arm antenna pattern vector

    Parameters
    ----------
    det_ten : np.ndarray
        The detector tensor.
    XX : np.ndarray
        The x-arm antenna pattern vector.
    YY : np.ndarray
        The y-arm antenna pattern vector.

    Returns
    -------
    dict
        The partial derivatives of the plus polarization of the antenna pattern.
    dict
        The partial derivatives of the cross polarization of the antenna pattern.
    '''
    # return [ [del_XX_Fp, del_YY_Fp], [del_XX_Fc, del_YY_Fc] ]
    return np.array([[np.matmul(det_ten,XX), -np.matmul(det_ten,YY)], [np.matmul(det_ten,YY), np.matmul(det_ten,XX)]])


# antenna pattern vectors
def calc_d_ant_pat_vectors(gra, dec, psi):
    '''
    Calculate the partial derivatives of the antenna pattern vectors with respect to the following parameters:
    - gra: the Greenwich right ascension
    - dec: the declination
    - psi: the polarization angle

    Parameters
    ----------
    gra : np.ndarray
        The Greenwich right ascension.
    dec : np.ndarray
        The declination.
    psi : np.ndarray
        The polarization angle.

    Returns
    -------
    np.ndarray
        The partial derivatives of the antenna pattern vector XX.
    np.ndarray
        The partial derivatives of the antenna pattern vector YY.
    '''
    # return [ [del_gra_XX, del_dec_XX, del_psi_XX],
    #          [del_gra_YY, del_dec_YY, del_psi_YY] ]
    return np.array([
        np.array([ [  cos(psi)*cos(gra) + sin(psi)*sin(gra)*sin(dec),  cos(psi)*sin(gra) - sin(psi)*cos(gra)*sin(dec),                   np.zeros_like(gra) ],
                   [                    - sin(psi)*cos(gra)*cos(dec),                    - sin(psi)*sin(gra)*cos(dec), -sin(psi)*sin(dec)*np.ones_like(gra) ],
                   [ -sin(psi)*sin(gra) - cos(psi)*cos(gra)*sin(dec),  sin(psi)*cos(gra) - cos(psi)*sin(gra)*sin(dec),  cos(psi)*cos(dec)*np.ones_like(gra) ]]),
       -np.array([ [ -sin(psi)*cos(gra) + cos(psi)*sin(gra)*sin(dec), -sin(psi)*sin(gra) - cos(psi)*cos(gra)*sin(dec),                   np.zeros_like(gra) ],
                   [                    - cos(psi)*cos(gra)*cos(dec),                    - cos(psi)*sin(gra)*cos(dec), -cos(psi)*sin(dec)*np.ones_like(gra) ],
                   [ -cos(psi)*sin(gra) + sin(psi)*cos(gra)*sin(dec),  cos(psi)*cos(gra) + sin(psi)*sin(gra)*sin(dec), -sin(psi)*cos(dec)*np.ones_like(gra) ]])])



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from .basic_relations import *
from .detector import Detector
from .injections import *
from .multi_network import MultiNetwork
from .network import Network
from .psd import noise_curves_built_in
from .waveform import Waveform
from .utils import det_specs_built_in

# available detector locations
available_locs = list(det_specs_built_in.keys())

# available detector technologies
available_tecs = list(noise_curves_built_in.keys())



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import gwbench.antenna_pattern_np as ant_pat_np

try:
    import gwbench.antenna_pattern_jx as ant_pat_jx
except ModuleNotFoundError:
    pass

def calc_wf_polarizations(net, f, inj_params):
    '''
    Calculate waveform polarizations directly from the passed frequency f and injection parameters
    inj_params and disregard whatever values are stored inside the Network object.

    Parameters
    ----------
    net : Network
        Network object.
    f : np.ndarray or jnp.ndarray
        Frequency array [Hz].
    inj_params : dict
        Injection parameters.

    Returns
    -------
    hfp : np.ndarray or jnp.ndarray
        Plus polarization.
    hfc : np.ndarray or jnp.ndarray
        Cross polarization.
    '''
    return net.wf.calc_wf_polarizations(f, inj_params)

def calc_det_responses(net, f, inj_params, trunc_f):
    '''
    Calculate detector responses.

    Calculate detector responses directly from the passed frequency f and injection parameters
    inj_params and disregard whatever values are stored inside the Network and Detector objects.

    Parameters
    ----------
    net : Network
        Network object.
    f : np.ndarray or jnp.ndarray
        Frequency array [Hz].
    inj_params : dict
        Injection parameters.
    trunc_f : bool
        Whether to truncate the frequency array of each detector according to its PSD setting.

    Returns
    -------
    hfs : list
        List of detector responses (either np.ndarray or jnp.ndarray).
    '''
    if net.wf.deriv_mod == 'jax': ant_pat = ant_pat_jx
    else:                         ant_pat = ant_pat_np

    if trunc_f: f_masks = [ (f >= det.f[0]) & (f <= det.f[-1]) for det in net.detectors ]
    else:       f_masks = len(net.detectors) * [ slice(None) ]

    hfp, hfc = calc_wf_polarizations(net, f, inj_params)

    return [ ant_pat.detector_response(hfp[f_mask], hfc[f_mask], f[f_mask],
                                       inj_params['Mc'], inj_params['eta'], inj_params['tc'],
                                       inj_params['ra'], inj_params['dec'], inj_params['psi'],
                                       det.loc, net.use_rot, user_locs=net.user_locs)
             for f_mask,det in zip(f_masks, net.detectors) ]



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


from functools import partial

import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as np

import gwbench.utils as utils
from gwbench.basic_relations import tau_spa_PN35

cos = np.cos
sin = np.sin
exp = np.exp
log = np.log
pi  = np.pi

i_vec = np.array([1,0,0])
j_vec = np.array([0,1,0])
k_vec = np.array([0,0,1])

def detector_response(hfp, hfc, f, Mc, eta, tc, ra, dec, psi, loc, use_rot, user_locs=None):
    '''
    Calculate the detector response for a given detector location and orientation.

    Parameters
    ----------
    hfp : jnp.ndarray
        The plus polarization.
    hfc : jnp.ndarray
        The cross polarization.
    f : jnp.ndarray
        The frequency domain array [Hz].
    Mc : float
        The chirp Mass [solar mass].
    eta: float
        The symmetric mass ratio.
    tc : float
        The time of coalescence [s].
    dec : float
        The declination [rad].
    ra : float
        The right ascension [rad].
    psi : float
        The polarization angle [rad].
    loc : str
        The location (and implied orientation) of a detector.
    use_rot : bool
        Use frequency dependent time due to rotation of earth and SPA.
    user_locs : dict, optional
        User defined locations and orientations of detectors.

    Returns
    -------
    hf : jnp.ndarray
        The detector response in the frequency domain.
    '''
    Fp, Fc, Flp = antenna_pattern_and_loc_phase_fac(f, Mc, eta, tc, ra, dec, psi, loc, use_rot, user_locs=user_locs)
    return Flp * (Fp * hfp + Fc * hfc)

def antenna_pattern_and_loc_phase_fac(f, Mc, eta, tc, ra, dec, psi, loc, use_rot, user_locs=None):
    '''
    Calculate the antenna pattern and location phase factor for a given detector location and orientation.

    Parameters
    ----------
    f : jnp.ndarray
        The frequency domain [Hz].
    Mc : float
        The chirp Mass [solar mass].
    eta: float
        The symmetric mass ratio.
    tc : float
        The time of coalescence [s].
    dec : float
        The declination [rad].
    ra : float
        The right ascencsion [rad].
    psi : float
        The polarization angle [rad].
    loc : str
        The location (and implied orientation) of a detector.
    use_rot : bool
        Use frequency dependent time due to rotation of earth and SPA.
    user_locs : dict, optional
        User defined locations and orientations of detectors.

    Returns
    -------
    Fp : jnp.ndarray
        The plus polarization antenna pattern.
    Fc : jnp.ndarray
        The cross polarization antenna pattern.
    Flp : jnp.ndarray
        The location phase factor.
    '''
    det_ten, det_vec, period_to_rad = det_quants(loc, user_locs=user_locs)
    time                            = calc_rotating_time(tc, f, Mc, eta, use_rot)
    time_delay                      = calc_time_delay(calc_gra(ra, time, period_to_rad), dec, det_vec)

    return *ant_pat_funcs(det_ten, *ant_pat_vectors(calc_gra(ra, time + time_delay, period_to_rad), dec, psi)), loc_phase_func(f, time_delay)

@partial(jax.jit, static_argnames=['use_rot'])
def calc_rotating_time(tc, f, Mc, eta, use_rot):
    '''
    Calculate the time including the effects of the rotation of the earth
    using the stationary phase approximation, if use_rot is True.

    Parameters
    ----------
    tc : float
        The time of coalescence [s].
    f : jnp.ndarray
        The frequency domain [Hz].
    Mc : float
        The chirp Mass [solar mass
    use_rot : bool
        Include rotation effects.

    Returns
    -------
    time : jnp.ndarray
        The time including rotation effects [s].
    '''
    if use_rot: return tc + utils.tc_offset - tau_spa_PN35(f, Mc, eta, log=log)
    else:       return np.array([tc + utils.tc_offset])

@jax.jit
def calc_gra(ra, time, period_to_rad):
    '''
    Calculate the Greenwich Right Ascension (GRA) for a given detector location and orientation.

    Parameters
    ----------
    ra : float
        The right ascencsion [rad].
    time : jnp.ndarray
        The corrected time [s].

    Returns
    -------
    gra : jnp.ndarray
        The Greenwich Right Ascension [rad].
    '''
    return ra - period_to_rad * time

@jax.jit
def calc_time_delay(gra, dec, det_vec):
    '''
    Calculate the time delay from the geocenter for a given detector location and orientation.

    Parameters
    ----------
    gra : jnp.ndarray
        Greenwich Right Ascension [rad].
    dec : float
        Declination [rad].
    det_vec : jnp.ndarray
        Detector location vector.

    Returns
    -------
    time_delay : jnp.ndarray
        Time delay from the geocenter [s].
    '''
    # using cos/sin(dec) instead of cos/sin(theta) with polar angle (theta = pi/2 - dec)
    return np.matmul(det_vec, np.array([cos(gra)*cos(dec), sin(gra)*cos(dec), sin(dec)*np.ones_like(gra)]))

@jax.jit
def loc_phase_func(f, time_delay):
    '''
    Calculate the location phase factor encoding the phase difference between the signal at the
    detector and the signal at the geocenter.

    Parameters
    ----------
    f : jnp.ndarray
        Frequency domain [Hz].
    time_delay : jnp.ndarray
        The corrected time [s].

    Returns
    -------
    Flp : jnp.ndarray
        Location phase factor
    '''
    return exp(1j * 2*pi * utils.mod_1(f * time_delay, np=np))

@jax.jit
def ant_pat_funcs(det_ten, XX, YY):
    '''
    Calculate the antenna pattern for a given detector location and orientation.

    Parameters
    ----------
    det_ten : jnp.ndarray
        The detector tensor.
    XX : jnp.ndarray
        The x-arm antenna pattern vector.
    YY : jnp.ndarray
        The y-arm antenna pattern vector.

    Returns
    -------
    Fp : jnp.ndarray
        The plus polarization antenna pattern.
    Fc : jnp.ndarray
        The cross polarization antenna pattern.
    '''
    return (0.5 * (np.matmul(det_ten,XX) * XX - np.matmul(det_ten,YY) * YY)).sum(axis=0), \
           (0.5 * (np.matmul(det_ten,XX) * YY + np.matmul(det_ten,YY) * XX)).sum(axis=0)

@jax.jit
def ant_pat_vectors(gra, dec, psi):
    '''
    Calculate the antenna pattern vectors for a given detector location and orientation.

    Parameters
    ----------
    gra : jnp.ndarray
        Greenwich Right Ascension [rad]
    dec : float
        Declination [rad]
    psi : float
        Polarization angle [rad]

    Returns
    -------
    XX : jnp.ndarray
        x-arm antenna pattern vector
    YY : jnp.ndarray
        y-arm antenna pattern vector
    '''
    return np.array([  cos(psi)*sin(gra) - sin(psi)*cos(gra)*sin(dec),
                      -cos(psi)*cos(gra) - sin(psi)*sin(gra)*sin(dec),
                                np.ones_like(gra) * sin(psi)*cos(dec) ]), \
           np.array([ -sin(psi)*sin(gra) - cos(psi)*cos(gra)*sin(dec),
                       sin(psi)*cos(gra) - cos(psi)*sin(gra)*sin(dec),
                                np.ones_like(gra) * cos(psi)*cos(dec) ])

def det_quants(loc, user_locs=None):
    '''
    Calculate the detector tensor, location vector, and period around geocenter for a given detector location and orientation.

    Parameters
    ----------
    loc : str
        Location (and implied orientation) of a detector.
    user_locs : dict, optional
        User defined locations and orientations of detectors.

    Returns
    -------
    det_ten : jnp.ndarray
        Detector tensor.
    det_vec : jnp.ndarray
        Detector location vector.
    period_to_rad : float
        Time to radian conversion factor for the passed period: 2pi/period [rad/s].
    '''
    alpha, beta, gamma, opening_angle, radius, period = utils.det_specs(loc, user_locs=user_locs)
    # insert polar angle theta = pi/2 - beta instead of latitude beta
    EulerD1 = np.matmul(np.matmul(rot_mat_2(alpha), rot_mat_1(pi/2 - beta)), rot_mat_2(gamma))

    eDArm1 = np.matmul(EulerD1, i_vec)
    eDArm2 = np.matmul(EulerD1, (cos(opening_angle) * i_vec + sin(opening_angle) * j_vec))

    return np.outer(eDArm1,eDArm1) - np.outer(eDArm2,eDArm2), -radius / utils.cLight * np.matmul(EulerD1, k_vec), 2*np.pi / period

@jax.jit
def rot_mat_1(angle):
    '''
    Calculate the rotation matrix for a given angle around axis 1.

    Parameters
    ----------
    angle : float
        Rotation angle [rad]

    Returns
    -------
    rot : jnp.ndarray
        Rotation matrix
    '''
    c = np.cos(angle)
    s = np.sin(angle)
    return np.array( [ [c,0,s], [0,1,0], [-s,0,c] ] )

@jax.jit
def rot_mat_2(angle):
    '''
    Calculate the rotation matrix for a given angle around axis 2.

    Parameters
    ----------
    angle : float
        Rotation angle [rad]

    Returns
    -------
    rot : jnp.ndarray
        Rotation matrix
    '''
    c = np.cos(angle)
    s = np.sin(angle)
    return np.array( [ [c,-s,0], [s,c,0], [0,0,1] ] )



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


'''This module handles calculations for a single gravitational wave detector.
'''

from copy import copy

import numpy as np

import gwbench.antenna_pattern_np as ant_pat_np
import gwbench.detector_response_derivatives as drd
import gwbench.err_deriv_handling as edh
import gwbench.fisher_analysis_tools as fat
import gwbench.psd as psd
import gwbench.snr as snr_mod
import gwbench.utils as utils

class Detector:
    """
    This class handles calculations for a single gravitational wave detector.

    Attributes
    ----------
    det_key : str
        full detector specification, specifying technology and location, e.g. CE2-40-CBO_C
    tec : str
        detector technology
    loc : str
        detector location
    f : np.ndarray
        frequency array
    psd : np.ndarray
        detector PSD
    Fp : np.ndarray
        plus polarization antenna pattern
    Fc : np.ndarray
        cross polarization antenna pattern
    Flp : np.ndarray
        location phase factor
    hf : np.ndarray
        detector response
    del_hf : dict
        derivative dictionary for detector responses
    del_hf_expr : dict
        sympy expression of derivative dictionary for detector responses
    snr : float
        SNR
    snr_sq : float
        SNR^2
    d_snr_sq : np.ndarray
        d(SNR^2) calculated from self.hf
    fisher : np.ndarray
        Fisher matrix
    cond_num : float
        condition number of Fisher matrix
    cov : np.ndarray
        covariance matrix
    inv_err : dict
        dictionary containing information about the inversion error between the two matrices
    errs : dict
        dictionary of errors for given derivative variables

    Methods
    -------
    __init__(det_key)
        Class constructor.
    set_f(f)
        Setter for frequency array.
    setup_psds(f_lo=-np.inf, f_hi=np.inf, user_psds=None)
        Sets up the detector PSD.
    setup_ant_pat_lpf(inj_params, use_rot, user_locs=None)
        Sets up the antenna pattern.
    calc_det_responses(wf, inj_params)
        Calculates the detector response.
    calc_det_responses_derivs_num(inj_params, wf, deriv_symbs_string, deriv_variables, conv_cos, conv_log, use_rot, step, method, order, d_order_n, user_locs, ana_deriv_symbs_string)
        Calculates the derivative dictionary for detector responses using numerical methods.
    load_det_responses_derivs_sym(wf_model_name, deriv_symbs_string, use_rot, gen_derivs=None, return_bin=0, user_lambdified_functions_path=None, logger=None)
        Loads the derivative dictionary for detector responses from a file.
    calc_det_responses_derivs_sym(wf, inj_params, deriv_symbs_string, deriv_variables, conv_cos, conv_log)
        Calculates the derivative dictionary for detector responses using sympy.
    calc_snrs(only_net, f_lo=-np.inf, f_hi=np.inf)
        Calculates the SNR and SNR^2.
    calc_snr_sq_integrand(f_lo=-np.inf, f_hi=np.inf)
        Calculates the integrand for SNR^2.
    calc_fisher_cov_matrices(only_net, cond_sup, f_lo=-np.inf, f_hi=np.inf, logger=None)
        Calculates the Fisher matrix and covariance matrix.
    calc_inv_err()
        Calculates the inversion error between the Fisher matrix and the covariance matrix.
    calc_errs(deriv_variables)
        Calculates the errors for given derivative variables.
    calc_sky_area_90(deriv_variables, logger=None)
        Calculates the 90% credible sky area.
    calc_sky_area_90_network(ra_id, dec_id, dec_val, is_cos_dec, dec_str)
        Calculates the 90% credible sky area for network.
    calc_cutler_vallisneri_overlap_vec(delta_hf=None, wf=None, inj_params=None, df=None, logger=None)
        Calculates the Cutler-Vallisneri overlap vector.
    calc_cutler_vallisneri_bias(logger=None)
        Calculates the Cutler-Vallisneri bias.
    print_detector(print_format=1)
        Prints the detector.
    """

    ###
    #-----Init methods-----
    def __init__(self, det_key):
        '''
        Class constructor.

        Parameters
        ----------
            det_key : str
                full detector specification, specifying technology and location, e.g. CE2-40-CBO_C
        '''
        #-----detector specification-----
        # full detector specification, specifying technology and location, e.g. CE2-40-CBO_C
        self.det_key = det_key
        # detector technology and locations
        self.tec = det_key.split('_')[0]
        self.loc = det_key.split('_')[1]

        #-----waveform and injection based quantities-----
        # frequency array
        self.f = None

        #-----technology and location based quantities-----
        # detector PSD
        self.psd = None
        # antenna pattern
        self.Fp = None
        self.Fc = None
        # location phase factor
        self.Flp = None

        #-----detector reponses-----
        # detector repsonse
        self.hf = None
        # derivative dictionary for detector responses
        self.del_hf = None
        # sympy expression of derivative dictionary for detector responses
        self.del_hf_expr = None

        #-----SNR-----
        # SNR, SNR^2 and d(SNR^2) calculated from self.hf
        self.snr = None
        self.snr_sq = None
        self.d_snr_sq = None

        #-----errors-----
        # Fisher matrix
        self.fisher = None
        # condition number of Fisher matrix
        self.cond_num = None
        # covariance matrix
        self.cov = None
        # dictionary containing information about the inversion error between the two matrices
        self.inv_err = None
        # dictionary of errors for given derivative variables
        self.errs = None
        # Cutler-Vallisneri bias quantities
        self.cutler_vallisneri_overlap_vec = None
        self.cutler_vallisneri_bias = None


    ###
    #-----Setter methods-----
    def set_f(self, f):
        '''
        Setter for frequency array.

        Parameters
        ----------
        f : np.ndarray
            frequency array
        '''
        self.f = copy(f)


    ###
    #-----PSDs and antenna patterns-----
    def setup_psds(self, f_lo=-np.inf, f_hi=np.inf, user_psds=None):
        '''
        Sets up the detector PSD.

        Parameters
        ----------
        f_lo : float
            lower frequency bound
        f_hi : float
            upper frequency bound
        user_psds : dict
            user-defined PSDs
        '''
        if user_psds is None or (self.det_key not in user_psds and self.tec not in user_psds):
            psd_file = None
            is_asd   = None
        elif self.det_key in user_psds:
            psd_file = user_psds[self.det_key]['psd_file']
            is_asd   = user_psds[self.det_key]['is_asd']
        else:
            psd_file = user_psds[self.tec]['psd_file']
            is_asd   = user_psds[self.tec]['is_asd']
        self.psd, self.f = psd.psd(self.tec,self.f,f_lo,f_hi,psd_file,is_asd)

    def setup_ant_pat_lpf(self, inj_params, use_rot, user_locs=None):
        '''
        Sets up the antenna pattern.

        Parameters
        ----------
        inj_params : dict
            injection parameters
        use_rot : bool
            use rotation
        user_locs : dict
            user-defined locations
        '''
        self.Fp, self.Fc, self.Flp = ant_pat_np.antenna_pattern_and_loc_phase_fac(self.f, inj_params.get('Mc'), inj_params.get('eta'),
            inj_params['tc'], inj_params['ra'], inj_params['dec'], inj_params['psi'], self.loc, use_rot, user_locs=user_locs)


    ###
    #-----Detector responses-----
    def calc_det_responses(self, wf, inj_params):
        '''
        Calculates the detector response.

        Parameters
        ----------
        wf : Waveform
            Waveform class instance
        inj_params : dict
            injection parameters

        Returns
        -------
        hf : np.ndarray
            detector response
        hfp : np.ndarray
            plus polarization of the waveform
        hfc : np.ndarray
            cross polarization of the waveform
        '''
        hfp, hfc = wf.calc_wf_polarizations(self.f, inj_params)
        self.hf = self.Flp * (hfp * self.Fp + hfc * self.Fc)
        return self.hf, hfp, hfc

    def calc_det_responses_derivs_num(self, inj_params, wf, deriv_symbs_string, deriv_variables, conv_cos, conv_log, use_rot,
                                      step, method, order, d_order_n, user_locs, ana_deriv_symbs_string):
        '''
        Calculates the partial derivatives of the detector response using numerical methods. The function also calculates
        the detector response.

        Parameters
        ----------
        inj_params : dict
            injection parameters
        wf : Waveform
            waveform
        deriv_symbs_string : str
            string of derivative symbols
        deriv_variables : list
            list of derivative variables
        conv_cos : bool
            convert to derivatives with respect to cosine of parameters
        conv_log : bool
            convert to derivatives with respect to logarithm of parameters
        use_rot : bool
            incorporate Earth's rotation
        step : float
            The step size for the numerical derivative.
        method : str
            The method for the numerical derivative.
        order : int
            The order of the numerical derivative.
        d_order_n : int
            The derivative order.
        user_locs : dict
            user-defined locations
        ana_deriv_symbs_string : str
            string of analytical derivative symbols
        '''
        _, hfp, hfc = self.calc_det_responses(wf, inj_params)
        self.del_hf = drd.calc_det_responses_derivs_num(self.loc, wf, deriv_symbs_string, self.f, inj_params, use_rot=use_rot, label='hf',
                                                        step=step, method=method, order=order, d_order_n=d_order_n, user_locs=user_locs,
                                                        ana_deriv_symbs_string=ana_deriv_symbs_string,
                                                        ana_deriv_aux={'hf':self.hf, 'hfp':hfp, 'hfc':hfc, 'Flp':self.Flp,
                                                                       'loc':self.loc, 'use_rot':use_rot, 'user_locs':user_locs})
        self.del_hf, c_quants = edh.get_conv_del_eval_dic(self.del_hf, inj_params, conv_cos, conv_log)
        inj_params, deriv_variables = edh.get_conv_inj_params_deriv_variables(c_quants, inj_params, deriv_variables)

    def load_det_responses_derivs_sym(self, wf_model_name, deriv_symbs_string, use_rot, gen_derivs=None, return_bin=0,
                                      user_lambdified_functions_path=None, logger=None):
        '''
        Loads the partial derivatives of the detector response.

        Parameters
        ----------
        wf_model_name : str
            waveform model name
        deriv_symbs_string : str
            string of derivative symbols
        use_rot : bool
            incorporate Earth's rotation
        gen_derivs : bool
            generate derivatives
        return_bin : int
            return binary
        user_lambdified_functions_path : str
            user lambdified functions path
        logger : Logger
            Python.logging logger instance
        '''
        self.del_hf_expr = drd.load_det_responses_derivs_sym(self.loc, wf_model_name, deriv_symbs_string, use_rot,
                                                             gen_derivs=gen_derivs, return_bin=return_bin, logger=logger,
                                                             user_lambdified_functions_path=user_lambdified_functions_path)

    def calc_det_responses_derivs_sym(self, wf, inj_params, deriv_symbs_string, deriv_variables, conv_cos, conv_log):
        '''
        Calculates the partial derivatives of the detector response using sympy. The function also calculates the detector response.

        Parameters
        ----------
        wf : Waveform
            waveform
        inj_params : dict
            injection parameters
        deriv_symbs_string : str
            string of derivative symbols
        deriv_variables : list
            list of derivative variables
        conv_cos : bool
            convert to derivatives with respect to cosine of parameters
        conv_log : bool
            convert to derivatives with respect to logarithm of parameters
        '''
        self.calc_det_responses(wf, inj_params)
        self.del_hf = {}
        for deriv in self.del_hf_expr:
            if deriv in ('variables', 'deriv_variables'): continue
            self.del_hf[deriv] = self.del_hf_expr[deriv](self.f, **utils.get_sub_dict(inj_params, self.del_hf_expr['variables']))

        self.del_hf, c_quants = edh.get_conv_del_eval_dic(self.del_hf, inj_params, conv_cos, conv_log)
        inj_params, deriv_variables = edh.get_conv_inj_params_deriv_variables(c_quants, inj_params, deriv_variables)


    ###
    #-----SNR calculations-----
    def calc_snrs(self, only_net, f_lo=-np.inf, f_hi=np.inf):
        '''
        Calculates the SNR and SNR^2.

        Parameters
        ----------
        only_net : bool
            if True only returns the SNR^2 value and does not store the SNR and SNR^2 values of the detector
        f_lo : float
            lower frequency bound, default is -np.inf
        f_hi : float
            upper frequency bound, default is np.inf

        Returns
        -------
        snr_sq : float
            SNR^2
        '''
        mask       = utils.min_max_mask(self.f, f_lo, f_hi)
        snr,snr_sq = snr_mod.snr_snr_square(self.hf[mask], self.psd[mask], self.f[mask])
        if not only_net:
            self.snr = snr
            self.snr_sq = snr_sq
        return snr_sq

    def calc_snr_sq_integrand(self, f_lo=-np.inf, f_hi=np.inf):
        '''
        Calculates the integrand for SNR^2.

        Parameters
        ----------
        f_lo : float
            lower frequency bound, default is -np.inf
        f_hi : float
            upper frequency bound, default is np.inf
        '''
        mask          = utils.min_max_mask(self.f, f_lo, f_hi)
        self.d_snr_sq = snr_mod.snr_square_integrand(self.hf[mask], self.psd[mask])


    ###
    #-----Error calculation and Fisher analysis-----
    def calc_fisher_cov_matrices(self, only_net, cond_sup, f_lo=-np.inf, f_hi=np.inf, logger=None):
        '''
        Calculates the Fisher matrix and covariance matrices.

        Parameters
        ----------
        only_net : bool
            if True only returns the Fisher matrix and does not store the Fisher matrix, covariance matrix, and inversion error of the detector
        cond_sup : float
            condition number support
        f_lo : float
            lower frequency bound, default is -np.inf
        f_hi : float
            upper frequency bound, default is np.inf
        logger : Logger
            Python.logging logger instance

        Returns
        -------
        fisher : np.ndarray
            Fisher matrix
        '''
        mask            = utils.min_max_mask(self.f, f_lo, f_hi)
        del_hf_sub_list = [ el[mask] for el in utils.get_sub_dict(self.del_hf, ['hf'], keep_in_dict=False).values() ]
        if not only_net:
            self.fisher, self.cov, self.cond_num, self.inv_err = fat.calc_fisher_cov_matrices(del_hf_sub_list, self.psd[mask], self.f[mask], only_fisher=0, cond_sup=cond_sup, logger=logger, tag=self.det_key)
            return self.fisher
        else:
            fisher,_,_,_ = fat.calc_fisher_cov_matrices(del_hf_sub_list, self.psd[mask], self.f[mask], only_fisher=1, cond_sup=cond_sup, logger=logger, tag=self.det_key)
            return fisher

    def calc_inv_err(self):
        '''
        Calculates the inversion error between the Fisher matrix and the covariance matrix.
        '''
        self.inv_err = fat.inv_err_from_fisher_cov(self.fisher, self.cov)

    def calc_errs(self, deriv_variables):
        '''
        Calculates the errors for given derivative variables.

        Parameters
        ----------
        deriv_variables : list
            list of derivative variables
        '''
        self.errs = fat.get_errs_from_cov(self.cov, deriv_variables)

    def calc_sky_area_90(self, deriv_variables, logger=None):
        '''
        Calculates the 90% credible sky area.

        Parameters
        ----------
        deriv_variables : list
            list of derivative variables
        logger : Logger
            Python.logging logger instance
        '''
        if self.cov is None or self.errs is None: return
        if 'ra' in deriv_variables and ('cos_dec' in deriv_variables or 'dec' in deriv_variables):
            if 'cos_dec' in deriv_variables: dec_str = 'cos_dec'
            else:                            dec_str = 'dec'
            ra_id      = deriv_variables.index('ra')
            dec_id     = deriv_variables.index(dec_str)
            is_cos_dec = (dec_str == 'cos_dec')
            self.errs['sky_area_90'] = edh.sky_area_90(self.errs['ra'],self.errs[dec_str],self.cov[ra_id,dec_id],self.inj_params['dec'],is_cos_dec)
        else: utils.log_msg(f'calc_sky_area_90: tag = {self.det_key} - Nothing done due to missing of either RA or COS_DEC (DEC) errors.', logger=logger, level='WARNING')

    def calc_sky_area_90_network(self, ra_id, dec_id, dec_val, is_cos_dec, dec_str, logger=None):
        '''
        Calculates the 90% credible sky area when called from an instance of Network (otherwise use calc_sky_area_90).

        Parameters
        ----------
        ra_id : int
            index of RA in the derivative variables
        dec_id : int
            index of DEC in the derivative variables
        dec_val : float
            declination value
        is_cos_dec : bool
            if True the declination is in cosine
        dec_str : str
            declination string
        '''
        if self.cov is None or self.errs is None: logger.warning(f'calc_sky_area_90_network: tag = {self.det_key} - 90%-credible sky area not calculated due to None-valued cov or errs.')
        else: self.errs['sky_area_90'] = edh.sky_area_90(self.errs['ra'], self.errs[dec_str], self.cov[ra_id,dec_id], dec_val, is_cos_dec)

    def calc_cutler_vallisneri_overlap_vec(self, delta_hf=None, wf=None, inj_params=None, df=None, logger=None):
        '''
        Calculates the Cutler-Vallisneri overlap vector, see Eq. (12) in Cutler and Vallisneri (2007).

        Parameters
        ----------
        delta_hf : np.ndarray
            difference between the detector responses
        wf : Waveform
            waveform model
        inj_params : dict
            injection parameters
        df : float
            frequency step
        logger : Logger
            Python.logging logger instance
        '''
        if delta_hf is None and (wf is None or inj_params is None):
            utils.log_msg(f'calc_cutler_vallisneri_overlap_vec: tag = {self.det_key} - Nothing done since delta_hf and wf/inj_params are None.', logger=logger, level='WARNING')
        else:
            if delta_hf is None:
                hfp, hfc = wf.eval_np_func(self.f, utils.get_sub_dict(inj_params, wf.wf_symbs_string))
                delta_hf = self.Flp * (hfp * self.Fp + hfc * self.Fc) - self.hf
            self.cutler_vallisneri_overlap_vec = snr_mod.cutler_vallisneri_overlap_vec(self.del_hf, delta_hf, self.psd, self.f, df=df)

    def calc_cutler_vallisneri_bias(self, logger=None):
        '''
        Calculates the Cutler-Vallisneri bias, see Eq. (12) in Cutler and Vallisneri (2007).

        Parameters
        ----------
        logger : Logger
            Python.logging logger instance
        '''
        if self.cov is None or self.cutler_vallisneri_overlap_vec is None:
            utils.log_msg(f'calc_cutler_vallisneri_bias: tag = {self.det_key} - Nothing done since cov or cutler_vallisneri_overlap_vec are None.',
                          logger=logger, level='WARNING')
        else: self.cutler_vallisneri_bias = snr_mod.cutler_vallisneri_bias(self.cov, self.cutler_vallisneri_overlap_vec)

    ###
    #-----IO methods-----
    def print_detector(self,print_format=1):
        '''Prints the detector.'''
        if print_format:
            sepl='-----------------------------------------------------------------------------------'
            print()
            print(sepl)
            print('Printing detector.')
            print(sepl)
            print()
        for key,value in vars(self).items():
            if type(value) == dict:
                print('Key: ',key)
                for key in value.keys():
                    print('',key)
                    print('',value[key])
                print()
            elif value is not None:
                print('Key: ',key)
                print(value)
                print()
        if print_format:
            print(sepl)
            print('Printing detector done.')
            print(sepl)
            print()



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


'''This module handles the benchmarking of graviational waveforms observed by multiple networks of detectors.

'''

import logging
from copy import copy, deepcopy

import dill
import numpy as np

import gwbench.network as nc
import gwbench.utils as utils

# logger
glob_logger = logging.getLogger('multi_network_module')
glob_logger.setLevel('INFO')

class MultiNetwork:
    '''
    Class for handling multiple networks of detectors and the benchmarking of gravitational waveforms observed by them.

    Attributes
    ----------
    network_specs : list
        list of network specifications, each of which is a list of detector keys.
    networks : list
        list of Network objects.
    loc_net : Network
        Network object for unique locations.

    Methods
    -------
    __init__(network_specs=None, logger_name='MultiNetwork', logger_level='WARNING', logger_level_network='WARNING')
        Initialize a MultiNetwork object.
    set_logger_level(level)
        Set the level of the logger.
    set_multi_network_from_specs(network_specs, logger_level_network='WARNING')
        Set the MultiNetwork object from the given network specifications.
    set_wf_vars(wf_model_name, wf_other_var_dic=None, user_waveform=None)
        Set the waveform variables for all networks.
    set_net_vars(wf_model_name=None, wf_other_var_dic=None, user_waveform=None,
                    f=None, inj_params=None, deriv_symbs_string=None, conv_cos=None, conv_log=None,
                    use_rot=None, user_locs=None, user_psds=None, user_lambdified_functions_path=None, ana_deriv_symbs_string=None)
        Set network variables for all networks (including the waveform variables).
    reset_ant_pat_lpf_psds()
        Reset the PSDs, antenna patterns, and LPFs for all networks.
    reset_wf_polarizations()
        Reset the waveform polarizations for all networks.
    reset_det_responses()
        Reset the detector responses for all networks.
    reset_snrs()
        Reset the SNRs for all networks.
    reset_errors()
        Reset the errors for all networks.
    check_none_vars_det(labels, ret_bool, tag='')
        Check for variables that are None in the networks.
    setup_ant_pat_lpf_psds(f_lo=-np.inf, f_hi=np.inf)
        Setup the PSDs, antenna patterns, and LPFs for all networks.
    calc_wf_polarizations()
        Calculate the waveform polarizations for the unique locations network.
    calc_wf_polarizations_derivs_num(step=1e-9, method='central', order=2, d_order_n=1)
        Calculate partial derivatives of the waveform polarizations numerically.
    load_wf_polarizations_derivs_sym(gen_derivs=None, return_bin=0)
        Load lambdified partial derivatives of the waveform polarizations for the unique locations network.
    calc_wf_polarizations_derivs_sym(gen_derivs=None)
        Calculate partial derivatives of the waveform polarizations using lambdified functions for the unique locations network.
    dist_wf_polarizations()
        Distribute the waveform polarizations and partial derivatives from the unique locations network to all networks.
    calc_det_responses()
        Calculate the detector responses for the unique locations network.
    calc_det_responses_derivs_num(step=1e-9, method='central', order=2, d_order_n=1)
        Calculate the partial derivatives of the detector responses numerically for the unique locations network.
    load_det_responses_derivs_sym(gen_derivs=None, return_bin=0)
        Load lambdified partial derivatives of the detector responses for the unique locations network.
    calc_det_responses_derivs_sym(gen_derivs=None, num_cores=None)
        Calculate the partial derivatives of the detector responses using lambdified functions for the unique locations network.
    dist_det_responses()
        Distribute the detector responses and partial derivatives from the unique locations network to all networks.
    calc_snrs(only_net=0)
        Calculate the SNRs for all networks.
    calc_snr_sq_integrand()
        Calculate the integrands of the SNR squared for all networks.
    calc_errors(cond_sup=None, only_net=0, derivs=None, step=None, method=None, order=None, gen_derivs=None, num_cores=None)
        Calculate the Fisher and covariance matrices and the errors for all networks.
    save_multi_network(filename_path)
        Save the network under the given path using *dill*.
    load_multi_network(filename_path)
        Loading the network from the given path using *dill*.
    '''

    ###
    #-----Init methods-----
    def __init__(self, network_specs=None, logger_name='MultiNetwork', logger_level='WARNING', logger_level_network='WARNING'):
        '''
        Initialize a MultiNetwork object.

        Parameters
        ----------
        network_specs : list
            list of network specifications, each of which is a list of detector keys.
        logger_name : str
            name of the logger.
        logger_level : str
            level of the logger.
        logger_level_network : str
            level of the logger for the networks.
        '''
        ##-----logger-----
        self.logger = utils.get_logger(name=logger_name, level=logger_level)

        ##-----initialize network object-----
        if network_specs is None:
            # network_specs list and list of networks
            self.network_specs = None
            self.networks = None
            # dummy networks for unique locations
            self.loc_net  = None
        elif isinstance(network_specs, list):
            self.set_multi_network_from_specs(network_specs, logger_level_network=logger_level_network)

        if self.networks is None: self.logger.debug('Empty MultiNetwork initialized.')
        else:                     self.logger.debug('MultiNetwork initialized.')

    ###
    #-----Setter methods-----
    #
    # it is best practice to always change the instance variables using these setter methods
    #
    def set_logger_level(self, level):
        '''
        Set the level of the logger.

        Parameters
        ----------
        level : str
            level of the logger.
        '''
        utils.set_logger_level(self.logger, level)

    def set_multi_network_from_specs(self, network_specs, logger_level_network='WARNING'):
        '''
        Set the MultiNetwork object from the given network specifications.

        Parameters
        ----------
        network_specs : list
            list of network specifications, each of which is a list of detector keys.
        logger_level_network : str
            level of the logger for the networks.
        '''
        # copy network_specs
        self.network_specs = copy(network_specs)
        # prepare list of all networks specified in network_specs
        self.networks = [ nc.Network(network_spec, logger_name='..'.join(network_spec), logger_level=logger_level_network)
                          for network_spec in network_specs ]
        # find unique locations and initialize Network for these with dummy technologies
        self.loc_net = nc.Network(
            list(dict.fromkeys([ f'tec_{det_key.split("_")[1]}' for network_spec in network_specs for det_key in network_spec ])),
            logger_name='loc_net', logger_level=logger_level_network)

    def set_wf_vars(self, wf_model_name, wf_other_var_dic=None, user_waveform=None):
        '''
        Set the waveform variables for all networks.

        Parameters
        ----------
        wf_model_name : str
            name of the waveform model.
        wf_other_var_dic : dict
            dictionary of other waveform variables.
        user_waveform : function
            user-defined waveform function.
        '''
        for net in [self.loc_net, *self.networks]:
            net.set_wf_vars(wf_model_name, wf_other_var_dic=wf_other_var_dic, user_waveform=user_waveform)

    def set_net_vars(self, wf_model_name=None, wf_other_var_dic=None, user_waveform=None,
                     f=None, inj_params=None, deriv_symbs_string=None, conv_cos=None, conv_log=None,
                     use_rot=None, user_locs=None, user_psds=None, user_lambdified_functions_path=None, ana_deriv_symbs_string=None):
        '''
        Set network variables for all networks (including the waveform variables). All variables are optional, with a default value of None.
        This is done to allow for a more flexible use of the method when some variables are already set.
        The variables that are strictly necessary are:

        Parameters
        ----------
        wf_model_name : str
            Name of the waveform model.
        wf_other_var_dic : dict, optional
            Dictionary of other waveform variables.
        user_waveform : function, optional
            User-defined waveform function.
        f : np.ndarray, optional
            Frequency array.
        inj_params : dict, optional
            Dictionary of injection parameters.
        deriv_symbs_string : str, optional
            String of derivative symbols.
        conv_cos : list, optional
            Flag for cosine conversion.
        conv_log : list, optional
            Flag for logarithmic conversion.
        use_rot : bool, optional
            Flag for using rotation.
        user_locs : dict, optional
            Dictionary of user-defined locations.
        user_psds : dict, optional
            Dictionary of user-defined PSDs.
        ana_deriv_symbs_string : str, optional
            String of analytical derivative symbols.
        user_lambdified_functions_path : str, optional
            Path to user-defined lambdified functions.
        '''
        for net in [self.loc_net, *self.networks]:
            net.set_net_vars(wf_model_name=wf_model_name, wf_other_var_dic=wf_other_var_dic, user_waveform=user_waveform,
                             f=f, inj_params=inj_params, deriv_symbs_string=deriv_symbs_string, conv_cos=conv_cos, conv_log=conv_log,
                             use_rot=use_rot, user_locs=user_locs, user_psds=user_psds, ana_deriv_symbs_string=ana_deriv_symbs_string,
                             user_lambdified_functions_path=user_lambdified_functions_path)


    ###
    #-----Resetter methods for instance variables-----
    def reset_ant_pat_lpf_psds(self):
        '''
        Reset the PSDs, antenna patterns, and LPFs for all networks.
        '''
        for net in [self.loc_net, *self.networks]:
            net.reset_ant_pat_lpf_psds()

    def reset_wf_polarizations(self):
        '''
        Reset the waveform polarizations for all networks.
        '''
        for net in [self.loc_net, *self.networks]:
            net.reset_wf_polarizations()

    def reset_det_responses(self):
        '''
        Reset the detector responses for all networks.
        '''
        for net in [self.loc_net, *self.networks]:
            net.reset_det_responses()

    def reset_snrs(self):
        '''
        Reset the SNRs for all networks.
        '''
        for net in [self.loc_net, *self.networks]:
            net.reset_snrs()

    def reset_errors(self):
        '''
        Reset the errors for all networks.
        '''
        for net in [self.loc_net, *self.networks]:
            net.reset_errors()


    ###
    #-----Helper-----
    def check_none_vars_det(self, labels, ret_bool, tag=''):
        '''
        Check for variables that are None in the networks.

        Parameters
        ----------
        labels : list
            List of variable names.
        ret_bool : bool
            Return boolean, if False log an error message, if True simply return True, if triggered.
        tag : str
            Tag for the log message.

        Returns
        -------
        bool
            Boolean, False if all variables are set, True if at least one variable is None and ret_bool is True.
        '''
        for net in [self.loc_net, *self.networks]:
            if net.check_none_vars_det(labels, ret_bool, tag=''): return True
        return False


    ###
    #-----PSDs and antenna patterns-----
    def setup_ant_pat_lpf_psds(self, f_lo=-np.inf, f_hi=np.inf):
        '''
        Setup the PSDs, antenna patterns, and LPFs for all networks.

        Parameters
        ----------
        f_lo : float
            Lower frequency limit.
        f_hi : float
            Upper frequency limit.
        '''
        for net in [self.loc_net, *self.networks]:
            net.setup_ant_pat_lpf_psds(f_lo=f_lo, f_hi=f_hi)
        self.logger.info('PSDs, antenna patterns, and LPFs loaded.')


    ###
    #-----Waveform polarizations-----
    def calc_wf_polarizations(self):
        '''
        Calculate the waveform polarizations for the unique locations network.
        '''
        self.loc_net.calc_wf_polarizations()

    def calc_wf_polarizations_derivs_num(self, step=1e-9, method='central', order=2, d_order_n=1):
        '''
        Calculate partial derivatives of the waveform polarizations numerically. The function also calculates the waveform polarizations.

        Parameters
        ----------
        step : float
            Step size for numerical differentiation.
        method : str
            Method for numerical differentiation.
        order : int
            Order of the numerical differentiation.
        d_order_n : int
            Number of derivatives.
        '''
        self.loc_net.calc_wf_polarizations_derivs_num(step=step, method=method, order=order, d_order_n=d_order_n)

    def load_wf_polarizations_derivs_sym(self, gen_derivs=None, return_bin=0):
        '''
        Load lambdified partial derivatives of the waveform polarizations for the unique locations network.

        Parameters
        ----------
        gen_derivs : bool
            Generate derivatives, if True. If False or None the derivatives are loaded from file, default is None.
        return_bin : int
            Return binary flag.
        '''
        self.loc_net.load_wf_polarizations_derivs_sym(gen_derivs=gen_derivs, return_bin=return_bin)

    def calc_wf_polarizations_derivs_sym(self, gen_derivs=None):
        '''
        Calculate partial derivatives of the waveform polarizations using lambdified functions for the unique locations network.
        The function also calculates the waveform polarizations.

        Parameters
        ----------
        gen_derivs : bool
            Generate derivatives, if True. If False or None the derivatives are loaded from file, default is None.
        '''
        self.loc_net.calc_wf_polarizations_derivs_sym(gen_derivs=gen_derivs)

    def dist_wf_polarizations(self):
        '''
        Distribute the waveform polarizations and partial derivatives from the unique locations network to all networks.
        '''
        for net in self.networks:
            net.hfp             =     copy(self.loc_net.hfp)
            net.hfc             =     copy(self.loc_net.hfc)
            net.del_hfpc        = deepcopy(self.loc_net.del_hfpc)
            net.del_hfpc_expr   = deepcopy(self.loc_net.del_hfpc_expr)
            net.inj_params      = deepcopy(self.loc_net.inj_params)
            net.deriv_variables =     copy(self.loc_net.deriv_variables)
        self.logger.info('Polarizations distributed among all networks.')


    ###
    #-----Detector responses-----
    def calc_det_responses(self):
        '''
        Calculate the detector responses for the unique locations network.
        '''
        self.loc_net.calc_det_responses()

    def calc_det_responses_derivs_num(self, step=1e-9, method='central', order=2, d_order_n=1, num_cores=None):
        '''
        Calculate the partial derivatives of the detector responses numerically for the unique locations network.

        Parameters
        ----------
        step : float
            Step size for numerical differentiation.
        method : str
            Method for numerical differentiation.
        order : int
            Order of the numerical differentiation.
        d_order_n : int
            Number of derivatives.
        num_cores : int
            Number of cores to use for parallel computation.
        '''
        self.loc_net.calc_det_responses_derivs_num(step=step, method=method, order=order, d_order_n=d_order_n, num_cores=num_cores)

    def load_det_responses_derivs_sym(self, gen_derivs=None, return_bin=0):
        '''
        Load lambdified partial derivatives of the detector responses for the unique locations network.

        Parameters
        ----------
        gen_derivs : bool
            Generate derivatives, if True. If False or None the derivatives are loaded from file, default is None.
        return_bin : int
            Return binary flag.
        '''
        self.loc_net.load_det_responses_derivs_sym(gen_derivs=gen_derivs, return_bin=return_bin)

    def calc_det_responses_derivs_sym(self, gen_derivs=None, num_cores=None):
        '''
        Calculate the partial derivatives of the detector responses using lambdified functions for the unique locations network.
        The function also calculates the detector responses.

        Parameters
        ----------
        gen_derivs : bool
            Generate derivatives, if True. If False or None the derivatives are loaded from file, default is None.
        num_cores : int
            Number of cores to use for parallel computation.
        '''
        self.loc_net.calc_det_responses_derivs_sym(gen_derivs=gen_derivs, num_cores=num_cores)

    def dist_det_responses(self):
        '''
        Distribute the detector responses and partial derivatives from the unique locations network to all networks.
        '''
        if self.check_none_vars_det(['psd'], True): self.setup_ant_pat_lpf_psds()
        for net in self.networks:
            net.inj_params          = deepcopy(self.loc_net.inj_params)
            net.deriv_variables     =     copy(self.loc_net.deriv_variables)
            for det in net.detectors:
                loc_det             = self.loc_net.get_detector(f'tec_{det.loc}')
                mask                = utils.min_max_mask(loc_det.f, det.f[0], det.f[-1])
                det.hf              = copy(loc_det.hf[mask])
                if loc_det.del_hf is not None:
                    det.del_hf = { deriv : copy(del_hf_deriv[mask]) for deriv,del_hf_deriv in loc_det.del_hf.items() }
                det.del_hf_expr     = deepcopy(loc_det.del_hf_expr)
        self.logger.info('Detector responses distributed among all networks.')


    ###
    #-----SNR calculations-----
    def calc_snrs(self, only_net=False, f_lo=-np.inf, f_hi=np.inf):
        '''
        Calculate the SNRs for all networks.

        Parameters
        ----------
        only_net : bool
            Calculate the SNRs for the networks only, default is False.
        '''
        if self.check_none_vars_det(['hf'], True):
            self.calc_det_responses()
            self.dist_det_responses()
        for net in self.networks:
            net.calc_snrs(only_net=only_net, f_lo=f_lo, f_hi=f_hi)
        self.logger.info('SNRs calculated.')

    def calc_snr_sq_integrand(self, f_lo=-np.inf, f_hi=np.inf):
        '''
        Calculate the integrands of the SNR squared for all networks.

        Parameters
        ----------
        only_net : bool
            Calculate the SNRs for the networks only, default is False.
        '''
        if self.check_none_vars_det(['hf'], True): self.calc_det_responses()
        for net in self.networks:
            net.calc_snr_sq_integrand(f_lo=f_lo, f_hi=f_hi)
        self.logger.info('SNR integrands calculated.')


    ###
    #-----Error calculation and Fisher analysis-----
    def calc_errors(self, cond_sup=None, only_net=0, derivs=None, step=None, method=None, order=None, gen_derivs=None, num_cores=None, f_lo=-np.inf, f_hi=np.inf):
        '''
        Calculate the Fisher and covariance matrices and the errors for all networks. If the derivatives are not pre-calculated,
        the method will calculate and distribute them using the specified method (derivs, step, method, order, gen_derivs, num_cores).

        Parameters
        ----------
        cond_sup : float
            Condition number suppression for the Fisher matrix.
        only_net : bool, optional
            Calculate the errors for the networks only, default is False.
        derivs : str, optional
            Type of derivatives, default is None. Valid options are: 'num' (numerical), 'sym' (symbolic).
        step : float, optional
            Step size for numerical differentiation, default is None.
        method : str, optional
            Method for numerical differentiation, default is None.
        order : int, optional
            Order of the numerical differentiation, default is None.
        gen_derivs : bool, optional
            Generate derivatives, if True, default is None.
        num_cores : int, optional
            Number of cores to use for parallel computation, default is None.
        '''
        if self.check_none_vars_det(['del_hf'], True):
            if not self.loc_net.check_none_vars_det(['del_hf'], True):
                self.dist_det_responses()
            elif derivs == 'num' and None not in [step, method, order]:
                self.calc_det_responses_derivs_num(step=step, method=method, order=order, d_order_n=1, num_cores=num_cores)
                self.dist_det_responses()
            elif derivs == 'sym':
                self.load_det_responses_derivs_sym(gen_derivs=gen_derivs)
                self.calc_det_responses_derivs_sym(num_cores=num_cores)
                self.dist_det_responses()
            else: utils.log_msg('calc_errors: Neither detector response derivatives have been pre-calculated, nor was ' +
                                'the differentiation method specified. Calculate the derivatives beforehand ' +
                                'or specify the derivative type (derivs=[num, sym]) and [step, method, order] for ' +
                                'the numerical differentiation.', logger=self.logger, level='ERROR')
        self.logger.info('Calculate errors (Fisher & cov matrices).')
        for net in self.networks:
            net.calc_errors(cond_sup=cond_sup, only_net=only_net, f_lo=f_lo, f_hi=f_hi)
        self.logger.info('Errors calculated.')


    ###
    #-----IO methods-----
    def save_multi_network(self, filename_path):
        '''Save the network under the given path using *dill*.'''
        with open(filename_path, "wb") as fi:
            dill.dump(self, fi, recurse=True)
        self.logger.info('MultiNetwork pickled.')
        return

    def load_multi_network(filename_path):
        '''Loading the network from the given path using *dill*.'''
        with open(filename_path, "rb") as fi:
            mul_net = dill.load(fi)
        mul_net.logger.info('MultiNetwork loaded.')
        return mul_net



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


from argparse import ArgumentParser

import numpy as np

from gwbench import MultiNetwork, injections_CBC_params_redshift, M_of_Mc_eta, f_isco_Msolar

np.set_printoptions(linewidth=200)

parser = ArgumentParser()
parser.add_argument('--inj', type = int, help = 'Injection ID in [0,100).', default = 0)
parser.add_argument('--derivs', type = str, help = 'Specify wich differentiation method to use: [num, sym].', default = 'num')

############################################################################
### User Choices
############################################################################

# choose between numeric or symbolic derivatives ['num', 'sym']
derivs = parser.parse_args().derivs

# choose injection id
inj_id = parser.parse_args().inj

# user's choice: waveform to use
wf_model_name = 'tf2'
#wf_model_name = 'tf2_tidal'
#wf_model_name = 'lal_bbh'
#wf_model_name = 'lal_bns'

# the lal_... waveform models are wrappers to call frequency domain waveform from lalsuite
# and thus need further specification of the approximant to use
if   wf_model_name == 'tf2':       wf_other_var_dic = None
elif wf_model_name == 'tf2_tidal': wf_other_var_dic = None
elif wf_model_name == 'lal_bbh':   wf_other_var_dic = {'approximant':'IMRPhenomD'}
elif wf_model_name == 'lal_bns':   wf_other_var_dic = {'approximant':'IMRPhenomD_NRTidalv2'}

# user defined waveform model, defined by the files specified in the dictionary
# only the 'np' one is needed when using numeric derivatives
if 1: user_waveform = None
else:
    wf_model_name    = 'tf2_user'
    wf_other_var_dic = None
    user_waveform   = {'np': '../gwbench/wf_models/tf2_np.py', 'sp':'../gwbench/wf_models/tf2_sp.py'}

# example detector location defined by user
user_locs = {'user-loc':{'longitude': 3.2, 'latitude': 0.4, 'arm_azimuth':0.3, 'which_arm':'y', 'shape':'L'}}
# example detector psd defined by the user
user_psds = {'user-tec':{'psd_file':'../gwbench/noise_curves/ce2_40km_cb.txt', 'is_asd':True}}

# user's choice: with respect to which parameters to take derivatives for the Fisher analysis
if 'tidal' in wf_model_name or 'bns' in wf_model_name: deriv_symbs_string = 'Mc eta chi1z chi2z DL tc phic iota lam_t ra dec psi'
else: deriv_symbs_string = 'Mc eta chi1z chi2z DL tc phic iota ra dec psi'

# user's choice: convert derivatives to cos or log for specific variables
conv_cos = ('dec','iota')
conv_log = ('Mc','DL','lam_t')

# if symbolic derivatives, take from generate_lambdified_functions.py
# if numeric  derivatives, user's decision
use_rot = 1
# 1 for True, 0 for False
# calculate SNRs, error matrices, and errors only for the network
only_net = 1

# number of cores to use for parallelize of the calc_det_responses_derivs
# = None for no parallelization, = 2,3,4,... to allocate N cores (even numbers preferred)
num_cores = None

# options for numeric derivative calculation
if derivs == 'num':
    # user's choice: switch particular partial derivatives to be analytical, options = [DL,tc,phic,ra,dec,psi]
    # otherwise set to None
    ana_deriv_symbs_string = 'DL tc phic ra dec psi'

    # choose numdifftools parameters for numerical derivatives
    step      = 1e-6
    method    = 'central'
    order     = 2

    # only relevant for symbolic derivatives
    gen_derivs = None

# options for symbolic derivative calculation
elif derivs == 'sym':

    # user's choice: switch particular partial derivatives to be analytical, options = [DL,tc,phic,ra,dec,psi]
    # otherwise set to None
    ana_deriv_symbs_string = None

    # choose numdifftools parameters for numerical derivatives
    step      = None
    method    = None
    order     = None

    # tell the code to generate symbolic derivatives as needed (turned on this tutorial)
    # the recommendation is to precompute them externally and load them for large-scale runs
    gen_derivs = True

# user's choice to generate injection parameters
if 'tidal' in wf_model_name or 'bns' in wf_model_name:
    mmin      = 0.8
    mmax      = 3
    chi_lo    = -0.05
    chi_hi    = 0.05
else:
    mmin      = 5
    mmax      = 100
    chi_lo    = -0.75
    chi_hi    = 0.75

cosmo_dict = {'zmin':0, 'zmax':0.2, 'sampler':'uniform_comoving_volume_inversion'}
mass_dict  = {'dist':'uniform', 'mmin':mmin, 'mmax':mmax}
# the default waveforms above are non-precessing, hence dim=1, set dim=3 for precessing waveforms like 'IMRPhenomPv2' or 'IMRPhenomPv2_NRTidalv2'
spin_dict  = {'dim':1, 'geom':'cartesian', 'chi_lo':chi_lo, 'chi_hi':chi_hi}

redshifted = 1
num_injs   = 100
seed       = 29378
file_path  = None

injections_data = injections_CBC_params_redshift(cosmo_dict,mass_dict,spin_dict,redshifted,num_injs,seed,file_path)

############################################################################
### injection parameters
############################################################################

inj_params = {
    'Mc'    : injections_data[0][inj_id],
    'eta'   : injections_data[1][inj_id],
    'chi1x' : injections_data[2][inj_id],
    'chi1y' : injections_data[3][inj_id],
    'chi1z' : injections_data[4][inj_id],
    'chi2x' : injections_data[5][inj_id],
    'chi2y' : injections_data[6][inj_id],
    'chi2z' : injections_data[7][inj_id],
    'DL'    : injections_data[8][inj_id],
    'tc'    : 0.,
    'phic'  : 0.,
    'iota'  : injections_data[9][inj_id],
    'ra'    : injections_data[10][inj_id],
    'dec'   : injections_data[11][inj_id],
    'psi'   : injections_data[12][inj_id],
    'z'     : injections_data[13][inj_id],
    }

if 'tidal' in wf_model_name or 'bns' in wf_model_name:
    inj_params['lam_t']       = 600.
    inj_params['delta_lam_t'] = 0.

print('injections parameter: ', inj_params)
print()

############################################################################
### Network specification
############################################################################

network_specs = [ ['CE2-40-CBO_H','CE2-40-CBO_L','CE2-40-CBO_V']        ,
                  ['CE2-40-CBO_H','CE2-40-CBO_L','CE2-40-CBO_user-loc'] ,
                  ['CE2-40-CBO_H','CE2-40-CBO_V','user-tec_user-loc']   ]
print('network specs: ', network_specs)
print()

f_lo = 1.
f_hi = f_isco_Msolar(M_of_Mc_eta(inj_params['Mc'],inj_params['eta']))
df   = 2.**-4
f    = np.arange(f_lo,f_hi+df,df)

print('f_lo:', f_lo, '   f_hi:', f_hi, '   df:', df)
print()

#############################################################################
### Multi Network GW Benchmarking
#############################################################################

# initialize MultiNetwork and do general setup
net = MultiNetwork(network_specs, logger_level='INFO', logger_level_network='WARNING')
# pass all the needed and optional variables
net.set_net_vars(wf_model_name=wf_model_name, wf_other_var_dic=wf_other_var_dic, user_waveform=user_waveform,
                 f=f, inj_params=inj_params, deriv_symbs_string=deriv_symbs_string,
                 conv_cos=conv_cos, conv_log=conv_log, use_rot=use_rot,
                 user_locs=user_locs, user_psds=user_psds, ana_deriv_symbs_string=ana_deriv_symbs_string)
# start the actual analysis
net.calc_errors(only_net=only_net, derivs=derivs, step=step, method=method, order=order, gen_derivs=gen_derivs, num_cores=num_cores)

### for completeness here are the steps involved, but since gwbench-0.7.4 the code will perform the necessary
### internally as needed
'''
net.setup_ant_pat_lpf_psds()
if derivs == 'num':
    net.calc_det_responses_derivs_num(step=step, method=method, order=order, num_cores=num_cores)
elif derivs = 'sym':
    # generation of lambdified derivatives used to be handles outside and is still recommended for speed reasons
    # refer to example_script/generate_lambdified_functions.py
    net.load_det_responses_derivs_sym(gen_derivs=True)
    net.calc_det_responses_derivs_sym(num_cores=num_cores)
# need to distribute the detector responses and derivatives within the MultiNetwork
net.dist_det_responses()
net.calc_snrs(only_net=only_net)
net.calc_errors(only_net=only_net)
'''

############################################################################
### Print results
############################################################################

print()
for subnet in net.networks:
    print(f'network = {subnet.label}')
    print(f'    snr = {subnet.snr}')
    print(f'   errs = {subnet.errs}')
    print()

print('Fisher analysis done.')



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


'''This module contains two methods that calculate numerical derivatives.
'''

import numdifftools as nd
import numpy
try:
    import jax
    jax.config.update("jax_enable_x64", True)
except ModuleNotFoundError:
    pass

import gwbench.analytic_derivatives as ad
import gwbench.utils as utils

def part_deriv_hf_func(hf, symbols_list, deriv_symbs_list, f, params_dic,
                       deriv_mod, pl_cr=True, compl=True, label='hf',
                       ana_deriv_symbs_list=[], ana_deriv_aux={},
                       step=1e-9, method='central', order=2, d_order_n=1):
    '''
    Calculate the partial derivatives of the waveform function numerically. The function can have one (hf) or two outputs (hfp, hfc), in the
    latter case the derivatives are calculated for the plus and cross polarizations the parameter [pl_cr=True] has to be passed. The derivatives
    are calculated with respect to the variables in [deriv_symbs_list] and evaluated at the values of the parameters in [params_dic].
    The numerical derivatives are calculated with the [numdifftools] package, while the diffentiation method and order can be set with the [step],
    [method], and [order] parameters. The order of the derivative can be set with the [d_order_n] parameter. Certain variables can be excluded from
    the derivatives by passing a list of their names to the [ana_deriv_symbs_list] parameter. The derivatives of the excluded variables are calculated
    analytically and the results are added to the numerical derivatives.

    Parameters
    ----------
    hf : function
        the waveform function. If it has two outputs, the first output is the plus polarization and the second output is the cross polarization.
    symbols_list : list
        the list of variables that the function depends on.
    deriv_symbs_list : list
        the list of variables for which the derivatives are calculated.
    f : float
        the frequency at which the waveform is evaluated.
    params_dic : dict
        the dictionary of parameters and their values at which the derivatives are evaluated.
    deriv_mod : str
        the method of numerical differentiation, choices are ['numdifftools', 'jax'].
    pl_cr : bool, optional
        if True, the derivatives are calculated for the plus and cross polarizations separately, default is True.
    compl : bool, optional
        if True, the function has two outputs and the derivatives are calculated for the real and imaginary parts separately, default is True.
    label : str, optional
        the label of waveform (hf, hfp, hfc), default is 'hf'.
    ana_deriv_symbs_list : list, optional
        the list of variables that are excluded from the numerical derivatives, default is None.
    ana_deriv_aux : dict, optional
        the dictionary of auxiliary variables that are used in the analytical derivatives, default is None.
    step : float, optional
        the step size for the numerical differentiation, default is 1e-9.
    method : str, optional
        the method of numerical differentiation, default is 'central'.
    order : int, optional
        the order of numerical differentiation, default is 2.
    d_order_n : int, optional
        the order of the derivative, default is 1.

    Returns
    -------
    del_hf_dic : dict
        dictionary that contains the numerical and analytical partial derivatives.
    '''

    label_keys            = int(pl_cr) * [label + 'p', label + 'c'] + int(not pl_cr) * [label]
    del_hf_dic_oder       = [ f'del_{name}_{label_key}' for name in deriv_symbs_list for label_key in label_keys ]
    _ana_deriv_symbs_list = [ el for el in deriv_symbs_list if el     in ana_deriv_symbs_list ]
    _num_deriv_symbs_list = [ el for el in deriv_symbs_list if el not in ana_deriv_symbs_list ]
    params_list           = [ params_dic[param] for param in symbols_list ]
    num_deriv_params_list = [ params_dic[param] for param in _num_deriv_symbs_list ]
    hf_ids                = { el : _num_deriv_symbs_list.index(el) for el in _num_deriv_symbs_list }

    if _ana_deriv_symbs_list:
        if pl_cr: del_hf_dic_ana = { f'del_{name}_{label_key}' : deriv for label_key in label_keys
                                     for name,deriv in ad.waveform_ana_derivs(f, params_dic, ana_deriv_aux[label_keys]).items() }
        else:     del_hf_dic_ana = { f'del_{name}_{label_key}' : deriv for label_key in label_keys
                                     for name,deriv in ad.detector_response_ana_derivs(f, params_dic, **ana_deriv_aux).items() }
    else:         del_hf_dic_ana = {}

    if not _num_deriv_symbs_list: return { key : del_hf_dic_ana[key] for key in del_hf_dic_oder }

    def hf_of_deriv_params(f, *deriv_params_list):
        return hf(f, *[deriv_params_list[hf_ids[el]] if el in _num_deriv_symbs_list else params_list[i] for i,el in enumerate(symbols_list)])

    del_hf = part_deriv(hf_of_deriv_params, f, num_deriv_params_list, deriv_mod, pl_cr=pl_cr, compl=compl, step=step, method=method, order=order, d_order_n=d_order_n)

    if not pl_cr: del_hf = [del_hf]

    del_hf_dic_num = { f'del_{name}_{label_key}' : _del_hf[:,i] for i,name in enumerate(_num_deriv_symbs_list) for label_key,_del_hf in zip(label_keys, del_hf) }

    return { key : { **del_hf_dic_num, **del_hf_dic_ana }[key] for key in del_hf_dic_oder }


def part_deriv(func, f, params_list, deriv_mod, pl_cr=0, compl=None, step=1e-9, method='central', order=2, d_order_n=1):
    '''
    Calculate the partial derivatives of a function numerically, see [part_deriv_gradient] for more details.

    Parameters
    ----------
    func : function
        The function for which the derivatives are calculated.
    f : float
        The frequency at which the function is evaluated.
    params_list : list, optional
        The list of parameters at which the derivatives are evaluated.
    deriv_mod : str
        The method of numerical differentiation. Choices are ['numdifftools', 'jax'].
    pl_cr : bool, optional
        If True, the derivatives are calculated for the two outputs of the function separately.
    compl : bool, optional
        If True, the function has two outputs and the derivatives are calculated for the real and imaginary parts separately.
    step : float, optional
        The step size for the numerical differentiation.
    method : str, optional
        The method of numerical differentiation.
    order : int, optional
        The order of numerical differentiation.
    d_order_n : int, optional
        The order of the derivative.

    Returns
    -------
    del_hf : array
        The numerical partial derivatives of the function.
    '''

    if deriv_mod == 'jax': np = jax.numpy
    else:                  np = numpy

    if pl_cr:
        if compl:
            def amp_pha_of_hfpc_compl(f, *params_list):
                return utils.pl_cr_to_amp_pha(*func(f, *params_list), np=np)

            amp_pl, pha_pl, amp_cr, pha_cr = amp_pha_of_hfpc_compl(f, *params_list)

            del_amp_pl = part_deriv_gradient(amp_pha_of_hfpc_compl, f, params_list, deriv_mod, 0, step, method, order, d_order_n)
            del_pha_pl = part_deriv_gradient(amp_pha_of_hfpc_compl, f, params_list, deriv_mod, 1, step, method, order, d_order_n)
            del_amp_cr = part_deriv_gradient(amp_pha_of_hfpc_compl, f, params_list, deriv_mod, 2, step, method, order, d_order_n)
            del_pha_cr = part_deriv_gradient(amp_pha_of_hfpc_compl, f, params_list, deriv_mod, 3, step, method, order, d_order_n)

            del_hfp = utils.z_deriv_from_amp_pha(amp_pl, pha_pl, del_amp_pl, del_pha_pl, np=np)
            del_hfc = utils.z_deriv_from_amp_pha(amp_cr, pha_cr, del_amp_cr, del_pha_cr, np=np)

            return del_hfp, del_hfc

        else:
            def amp_pha_of_hfpc_real(f, *params_list):
                return utils.amp_pha_from_re_im(*func(f, *params_list), np=np)

            amp, pha = amp_pha_of_hfpc_real(f, *params_list)
            del_amp = part_deriv_gradient(amp_pha_of_hfpc_real, f, params_list, deriv_mod, 0, step, method, order, d_order_n)
            del_pha = part_deriv_gradient(amp_pha_of_hfpc_real, f, params_list, deriv_mod, 1, step, method, order, d_order_n)

            return utils.re_im_from_z(utils.z_deriv_from_amp_pha(amp, pha, del_amp, del_pha, np=np), np=np)

    else:
        if compl:
            def amp_pha_of_hf(f, *params_list):
                return utils.amp_pha_from_z(func(f, *params_list), np=np)

            amp, pha = amp_pha_of_hf(f, *params_list)
            del_amp = part_deriv_gradient(amp_pha_of_hf, f, params_list, deriv_mod, 0, step, method, order, d_order_n)
            del_pha = part_deriv_gradient(amp_pha_of_hf, f, params_list, deriv_mod, 1, step, method, order, d_order_n)

            return utils.z_deriv_from_amp_pha(amp, pha, del_amp, del_pha, np=np)

        else:
            return part_deriv_gradient(func, f, params_list, deriv_mod, None, step, method, order, d_order_n)


def part_deriv_gradient(func, f, params_list, deriv_mod, funcid=None, step=1e-9, method='central', order=2, d_order_n=1):
    '''
    A wrapper for various gradient derivative methods:
      - [numdifftools.Gradient] - can be used with lalsimulation and tf2 waveforms.
      - [jax.jacrev] - ca be used with jax waveforms (tf2_jx, tf2_tidal_jx).
        The packages ripple and wf4py could be used, but require wrappers to be written.

    Parameters
    ----------
    func : function
        The function for which the derivatives are calculated.
    f : float
        The frequency at which the function is evaluated.
    params_list : list
        The list of parameters at which the derivatives are evaluated.
    deriv_mod : str
        The method of numerical differentiation. Choices are ['numdifftools', 'jax'].
    funcid : int, optional
        The index of the output of the function for which the derivatives are calculated.
    step : float, optional
        The step size for the numerical differentiation.
    method : str, optional
        The method of numerical differentiation.
    order : int, optional
        The order of numerical differentiation.
    d_order_n : int, optional
        The order of the derivative.

    Returns
    -------
    del_hf : array
        The numerical partial derivatives of the function.
    '''

    if deriv_mod == 'numdifftools':

        if funcid is None:
            def wraps(x):
                return func(f, *x)
        else:
            def wraps(x):
                return func(f, *x)[funcid]

        if len(params_list) == 1: return nd.Gradient(wraps, step=step, method=method, order=order, n=d_order_n)(params_list)[:,None]
        else:                     return nd.Gradient(wraps, step=step, method=method, order=order, n=d_order_n)(params_list)

    elif deriv_mod == 'jax':

        if funcid is None:
            def wraps(*x):
                return func(f, *x)
        else:
            def wraps(*x):
                return func(f, *x)[funcid]

        return numpy.array(jax.jacrev(wraps, argnums=range(len(params_list)))(*params_list)).T

    else: raise ValueError(f'part_deriv_gradient: The derivative module {deriv_mod} is not supported.')



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import numpy as np
from scipy.integrate import simpson

#-----overlap function-------
def scalar_product_integrand(hf, gf, psd):
    '''
    Calculate the integrand of the scalar product given two frequency domain waveforms and the power spectral density.

    Parameters
    ----------
    hf : np.ndarray
        Frequency domain waveform.
    gf : np.ndarray
        Frequency domain waveform.
    psd : np.ndarray
        Power spectral density.

    Returns
    -------
    np.ndarray
        Integrand of the scalar product.
    '''
    return 4 * (hf.real * gf.real + hf.imag * gf.imag) / psd

def scalar_product(hf, gf, psd, freqs):
    """
    Calculate the scalar product given two frequency domain waveforms, the power spectral density, and the frequency array.

    Parameters
    ----------
    hf : np.ndarray
        Frequency domain waveform.
    gf : np.ndarray
        Frequency domain waveform.
    psd : np.ndarray
        Power spectral density.
    freqs : np.ndarray
        Frequency array.

    Returns
    -------
    float
        Scalar product.
    """
    return simpson(scalar_product_integrand(hf, gf, psd), x=freqs)

#-----SNR function-------
def snr_square_integrand(hf, psd):
    """
    Calculate the integrand of the square of the signal-to-noise ratio given a frequency domain waveform and the power spectral density.

    Parameters
    ----------
    hf : np.ndarray
        Frequency domain waveform.
    psd : np.ndarray
        Power spectral density.

    Returns
    -------
    np.ndarray
        Integrand of the square of the signal-to-noise ratio.
    """
    return scalar_product_integrand(hf, hf, psd)

def snr_square(hf, psd, freqs):
    """
    Calculate the square of the signal-to-noise ratio given a frequency domain waveform, the power spectral density, and the frequency array.

    Parameters
    ----------
    hf : np.ndarray
        Frequency domain waveform.
    psd : np.ndarray
        Power spectral density.
    freqs : np.ndarray
        Frequency array.

    Returns
    -------
    float
        Square of the signal-to-noise ratio.
    """
    return scalar_product(hf, hf, psd, freqs)

def snr(hf, psd, freqs):
    """
    Calculate the signal-to-noise ratio given a frequency domain waveform, the power spectral density, and the frequency array.

    Parameters
    ----------
    hf : np.ndarray
        Frequency domain waveform.
    psd : np.ndarray
        Power spectral density.
    freqs : np.ndarray
        Frequency array.

    Returns
    -------
    float
        Signal-to-noise ratio.
    """
    return np.sqrt(snr_square(hf, psd, freqs))

def snr_snr_square(hf, psd, freqs):
    """
    Calculate the signal-to-noise ratio and the square of the signal-to-noise ratio given a frequency domain waveform, the power spectral density, and the frequency array.

    Parameters
    ----------
    hf : np.ndarray
        Frequency domain waveform.
    psd : np.ndarray
        Power spectral density.
    freqs : np.ndarray
        Frequency array.

    Returns
    -------
    float
        Signal-to-noise ratio.
    float
        Square of the signal-to-noise ratio.
    """
    snr_sq = snr_square(hf, psd, freqs)
    return np.sqrt(snr_sq), snr_sq

#-----Cutler Vallisneri bias-----
def cutler_vallisneri_overlap_vec(del_hf, delta_hf, psd, freqs):
    """
    Calculate the Cutler-Vallisneri overlap vector, see Eq. (12) in Cutler and Vallisneri (2007).

    Parameters
    ----------
    del_hf : np.ndarray
        Partial derivatives of the frequency domain waveform.
    delta_hf : np.ndarray
         difference between the detector responses in the frequency domain.
    psd : np.ndarray
        Power spectral density.
    freqs : np.ndarray
        Frequency array.

    Returns
    -------
    np.ndarray
        Cutler-Vallisneri overlap vector.
    """
    return np.array([ scalar_product(del_hf[deriv], delta_hf, psd, freqs) for deriv in del_hf ])

def cutler_vallisneri_bias(cov, overlap_vec):
    """
    Calculate the Cutler-Vallisneri bias given the covariance matrix and the overlap vector, see Eq. (12) in Cutler and Vallisneri (2007).

    Parameters
    ----------
    cov : np.ndarray
        Covariance matrix.
    overlap_vec : np.ndarray
        Overlap vector.

    Returns
    -------
    np.ndarray
        Cutler-Vallisneri bias.
    """
    return np.matmul(cov, overlap_vec)

#-----fft method from Anuradha-------
def rfft_normalized(time_series, dt, n=None):
    """
    Calculate the normalized real-valued discrete Fourier transform of a time series.

    Parameters
    ----------
    time_series : np.ndarray
        Time series.
    dt : float
        Time step.
    n : int, optional
        Number of points in the output.

    Returns
    -------
    np.ndarray
        Normalized real-valued discrete Fourier transform.
    """
    return np.fft.rfft(time_series, n) * dt

def fft_normalized(time_series, dt, n=None):
    """
    Calculate the normalized complex-valued discrete Fourier transform of a time series.

    Parameters
    ----------
    time_series : np.ndarray
        Time series.
    dt : float
        Time step.
    n : int, optional
        Number of points in the output.

    Returns
    -------
    np.ndarray
        Normalized complex-valued discrete Fourier transform.
    """
    return np.fft.fft(time_series, n) * dt



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import mpmath as mp
import numpy as np

import gwbench.snr as snr_mod
from gwbench.utils import log_msg

def calc_fisher_cov_matrices(del_hf_list, psd, f, only_fisher=False, cond_sup=None, fac=101, logger=None, tag=''):
    '''
    Calculate the Fisher matrix and its inverse covariance matrix for a list of waveform derivatives.

    Parameters
    ----------
    del_hf_list : list
        list of waveform derivatives
    psd : np.ndarray
        power spectral density
    f : np.ndarray
        frequency array
    only_fisher : bool, optional
        if True, only the Fisher matrix is calculated, default is False.
    cond_sup : float, optional
        condition number threshold for using the high precision mpmath library, default is None.
    fac : float, optional
        factor for the condition number threshold, default is 101.
    logger : logging.Logger, optional
        logger object, default is None.
    tag : str, optional
        tag for the logger, default is an empty string.

    Returns
    -------
    fisher : np.ndarray
        Fisher matrix
    cov : np.ndarray
        covariance matrix
    cond_num : float
        condition number of the Fisher matrix
    inv_err : dict
        dictionary containing information regarding the quality of the inversion
    '''
    n = len(del_hf_list)
    fisher = np.zeros((n,n))

    for i in np.arange(n):
        fisher[i,i]     = snr_mod.scalar_product(del_hf_list[i], del_hf_list[i], psd, f)
        for j in np.arange(i+1,n):
            fisher[i,j] = snr_mod.scalar_product(del_hf_list[i], del_hf_list[j], psd, f)
            fisher[j,i] = fisher[i,j]

    if only_fisher: return fisher, None, None, None

    cond_num     = calc_cond_number(fisher, logger=logger, tag=tag)
    cov, inv_err = calc_cov_inv_err(fisher, cond_num, cond_sup, fac, logger=logger, tag=tag)
    return fisher, cov, cond_num, inv_err

def calc_cond_number(fisher, logger=None, tag=''):
    '''
    Calculate the condition number of the Fisher matrix.

    Parameters
    ----------
    fisher : np.ndarray
        Fisher matrix
    logger : logging.Logger, optional
        logger object, default is None.
    tag : str, optional
        tag for the logger, default is an empty string.

    Returns
    -------
    cond_num : float
        condition number of the Fisher matrix
    '''
    try:
        EWs,_ = np.linalg.eig(fisher)
    except np.linalg.LinAlgError:
        log_msg(f'calc_cond_number: tag = {tag} - Eigenvalue calculation failed due to a LinAlgError!', logger=logger, level='WARNING')
        return np.nan

    if not np.amin(np.abs(EWs)):
        log_msg(f'calc_cond_number: tag = {tag} - Fisher matrix is singular!', logger=logger, level='WARNING')
        return np.inf
    else:
        return np.amax(np.abs(EWs))/np.amin(np.abs(EWs))

def calc_cov_inv_err(fisher, cond_num=None, cond_sup=None, fac=101, logger=None, tag=''):
    '''
    Calculate the covariance matrix and the inversion error from the Fisher matrix.

    Parameters
    ----------
    fisher : np.ndarray
        Fisher matrix
    cond_num : float, optional
        condition number of the Fisher matrix, default is None. 
    cond_sup : float, optional
        condition number threshold for using the high precision mpmath library, default is None.
    fac : float, optional
        factor for the condition number threshold, default is 101.
    logger : logging.Logger, optional
        logger object, default is None.
    tag : str, optional
        tag for the logger, default is an empty string.

    Returns
    -------
    cov : np.ndarray
        covariance matrix
    inv_err : dict
        dictionary containing information regarding the quality of the inversion
    '''
    if cond_num is None:             cond_num = calc_cond_number(fisher, logger=logger, tag=tag)
    if cond_num in [np.inf, np.nan]: return None, None # Singular Fisher matrix or LinAlgError

    try:
        if cond_sup is not None and cond_sup < 1e16 and cond_sup > cond_num:
            cov   = np.linalg.inv(fisher)
            cov   = (cov + cov.T) / 2
            ident = np.identity(len(fisher))
            return cov, inv_err_from_fisher_cov(fisher, cov, ident, fac, None)
        else:
            mp.mp.dps = max(int(np.ceil(np.log10(cond_num))) + 2, 20)
            fisher_mp = mp.matrix(fisher)
            cov_mp    = fisher_mp**-1
            cov_mp    = (cov_mp + cov_mp.T) / 2
            ident_mp  = mp.matrix(np.identity(len(fisher)))
            return conv_mp_to_np(cov_mp), inv_err_from_fisher_cov(fisher_mp, cov_mp, ident_mp, fac, mp.mp.dps)

    except (ZeroDivisionError, np.linalg.LinAlgError):
        log_msg(f'calc_cov_inv_err: tag = {tag} - There was an exception in the inversion of the Fisher matrix!',
                logger=logger, level='WARNING')
        return None, None

def get_errs_from_cov(cov, deriv_variables):
    '''
    Pull the standard deviations from the covariance matrix.

    Parameters
    ----------
    cov : np.ndarray
        covariance matrix
    deriv_variables : list
        list of variable names

    Returns
    -------
    errs : dict
        dictionary containing the standard deviations of the variables
    '''
    if cov is None:              return None
    if np.any(np.diag(cov) < 0): return { name : np.nan for name in deriv_variables }
    else:                        return { name : np.sqrt(cov[i,i]) for i,name in enumerate(deriv_variables) }

def conv_mp_to_np(matrix, dtype=np.float32):
    '''
    Convert a matrix from the mpmath library to a numpy array.

    Parameters
    ----------
    matrix : mpmath.matrix
        matrix to be converted
    dtype : type, optional
        data type of the numpy array, default is np.float32.

    Returns
    -------
    np.matrix
        numpy array
    '''
    return np.matrix(matrix.tolist(), dtype=dtype)

def inv_err_from_fisher_cov(fisher, cov, ident, fac, dps):
    '''
    Calculate the inversion error from the Fisher matrix and the covariance matrix.

    Parameters
    ----------
    fisher : np.ndarray
        Fisher matrix
    cov : np.ndarray
        covariance matrix
    ident : np.ndarray
        identity matrix
    fac : float
        factor for the condition number threshold
    dps : int
        decimal precision for the mpmath library

    Returns
    -------
    inv_err : dict
        dictionary containing information regarding the quality of the inversion
    '''
    cov_err, delta = calc_cov_err_inv_res(fisher, cov, ident)
    if dps is None:
        inv_err                = {'fisher_mp': None,   'cov_mp': None, 'dps_mp':None}
        inv_err['inv_err_mat'] = delta
        inv_err['inv_err']     = np.amax(np.abs(delta))
        inv_err['inv_cond']    = np.all(np.abs(fac * cov_err) < np.abs(cov))
        return inv_err
    else:
        inv_err                = {'fisher_mp': fisher, 'cov_mp': cov,  'dps_mp':dps}
        inv_err['inv_err_mat'] = delta
        inv_err['inv_err']     = np.amax(np.abs(conv_mp_to_np(delta)))
        inv_err['inv_cond']    = np.all(np.abs(fac * conv_mp_to_np(cov_err)) < np.abs(conv_mp_to_np(cov)))
        return inv_err

def calc_cov_err_inv_res(fisher, cov, ident):
    '''
    Calculate the covariance error and the inversion residual from the Fisher matrix and the covariance matrix.

    Parameters
    ----------
    fisher : np.ndarray
        Fisher matrix
    cov : np.ndarray
        covariance matrix
    ident : np.ndarray
        identity matrix

    Returns
    -------
    cov_err : np.ndarray
        covariance error
    delta : np.ndarray
        inversion residual

    Notes
    -----
    The inversion residual is calculated as follows:
    delta = fisher * cov - ident
    The covariance error is calculated as follows:
    cov_err = cov * delta * (ident + delta)**-1
    '''
    if isinstance(fisher, np.ndarray):
        delta = np.matmul(fisher, cov) - ident
        return np.matmul(np.matmul(cov, delta), np.linalg.inv(ident + delta)), delta
    else:
        delta = fisher * cov - ident
        return cov * delta * (ident + delta)**-1, delta
